"""
Analytic expression for Laguerre-Gaussian modes.

----

.. [1] L. Allen et al. "Orbital angular momentum of light and the transformation of 
    Laguerre-Gaussian laser modes." PRA 45.11 (1992): 8185.
"""
from math import factorial

import numexpr as ne
import numpy as np
from scipy.constants import c, pi
from scipy.special import genlaguerre

from quvac.field.abc import ExplicitField


class LaguerreGaussianAnalytic(ExplicitField):
    """
    Analytic expression for Laguerre-Gaussian modes.

    Parameters
    ----------
    field_params : dict
        Dictionary containing the field parameters. Required keys are:
            - 'focus_x' : tuple of float
                Location of spatial focus (x, y, z).
            - 'focus_t' : float
                Location of temporal focus.
            - 'theta' : float
                Polar angle of k-vector (in degrees).
            - 'phi' : float
                Azimuthal angle of k-vector (in degrees).
            - 'beta' : float
                Polarization angle (in degrees).
            - 'lam' : float
                Wavelength of the pulse.
            - 'w0' : float
                Waist size.
            - 'tau' : float
                Duration.
            - 'phase0' : float
                Phase delay at focus.
            - 'p' : int
                Radial index of the Laguerre-Gaussian mode.
            - 'l' : int
                Azimuthal index of the Laguerre-Gaussian mode.
            - 'E0' : float, optional
                Amplitude (either E0 or W is required).
            - 'W' : float, optional
                Energy (either E0 or W is required).
    grid : quvac.grid.GridXYZ
        Spatial and spectral grid.

    Notes
    -----
    All field parameters are in SI units.

    Analytic expression is from [1]_.
    """

    def __init__(self, field_params, grid):
        super().__init__(field_params, grid)

        self.phase0 += pi

        if "E0" not in field_params:
            err_msg = ("Field params need to have either W (energy) or"
                       "E0 (amplitude) as key")
            assert "W" in field_params, err_msg
            self.E0 = 1.0e10

        # Define additional field variables
        self.x0, self.y0, self.z0 = self.focus_x
        self.t0 = self.focus_t
        self.B0 = self.E0 / c
        self.k = 2.0 * pi / self.lam
        self.omega = c * self.k
        self.zR = pi * self.w0**2 / self.lam
        p, l_phi = self.p, self.l
        self.prefactor = np.sqrt(2*factorial(p)/factorial(p+l_phi)/pi)

        # Rotate coordinate grid
        self.rotate_coordinates()

        # Define variables not depending on time step
        self.w = "(w0 * sqrt(1. + (z/zR)**2))"
        self.r = "sqrt(x**2 + y**2)"
        self.r2 = "(x**2 + y**2)"
        self.R_inv = "(z/(z**2 + zR**2))"
        
        self.rw = ne.evaluate(f"({self.r}*sqrt(2)/{self.w})", global_dict=self.__dict__)
        self.lag_poly = self.rw**l_phi * genlaguerre(p, l_phi)(self.rw**2)

        self.E_expr = (f"B0 * w0/{self.w} * exp(-{self.r2}/{self.w}**2) * lag_poly")
        self.phase_no_t = ne.evaluate(
            (f"phase0 - k*{self.r2}*{self.R_inv}/2 + (2*p+l+1)*arctan(z/zR) + "
             "l*arctan2(y,x)"),
            global_dict=self.__dict__,
        )

        self.E = ne.evaluate(self.E_expr, global_dict=self.__dict__)

        # Set up correct field amplitude
        if "W" in field_params:
            self.check_energy()

    def check_energy(self):
        """
        Check and adjust the field energy.
        """
        self._check_energy()

        if self.modify_energy:
            self.E0 *= self.W_correction
            self.B0 = self.E0 / c
            self.E = ne.evaluate(self.E_expr, global_dict=self.__dict__)

    def calculate_field(self, t, E_out=None, B_out=None, mode="real"):
        """
        Calculates the electric and magnetic fields at a given time step.
        """
        k = 2.0 * pi / self.lam # noqa: F841
        self.psi_plane = ne.evaluate("(omega*(t-t0) - k*z)", global_dict=self.__dict__)
        self.phase = "(phase_no_t + psi_plane)"

        Et = ne.evaluate(
            f"E * exp(-(2.*psi_plane/(omega*tau))**2) * exp(-1.j*{self.phase})",
            global_dict=self.__dict__,
        )

        self.Ex = self.By = Et.copy()
        self.Ey = self.Ez = self.Bx = self.Bz = 0.0

        if mode == "real":
            self.convert_fields_to_real()

        E_out, B_out = self.rotate_fields_back(E_out, B_out, mode)
        return E_out, B_out



"""
Utility functions related to fields.
"""

import numexpr as ne
import numpy as np
from scipy.constants import c, epsilon_0, mu_0, pi


def get_field_energy(E, B, dV):
    """
    Calculate the numerical energy of the field on a spatial grid.

    Parameters
    ----------
    E : np.array
        Electric field components as a tuple of three numpy arrays (Ex, Ey, Ez).
    B : np.array
        Magnetic field components as a tuple of three numpy arrays (Bx, By, Bz).
    dV : float
        Volume element.

    Returns
    -------
    float
        The calculated energy of the field.

    Notes
    -----
    The energy is calculated using the formula:

    W = sum(0.5 * epsilon_0 * c**2 * dV * abs(E)**2 + 0.5 / mu_0 * dV * abs(B)**2)
    """
    Ex, Ey, Ez = E
    Bx, By, Bz = B
    W = 0.5 * epsilon_0 * c**2 * dV * ne.evaluate("sum(Ex**2 + Ey**2 + Ez**2)")
    W += 0.5 / mu_0 * dV * ne.evaluate("sum(Bx**2 + By**2 + Bz**2)")
    return W


def get_field_energy_kspace(a1, a2, k, dVk, mode="without 1/k"):
    """
    Calculate the numerical energy of the field on a spectral grid.

    Parameters
    ----------
    a1 : np.array
        First spectral coefficient.
    a2 : np.array
        Second spectral coefficient.
    k : np.array
        Wavevector.
    dVk : float
        Spectral volume element.
    mode : str, optional
        Convention for the definition of a1 and a2 ('without 1/k' or 'with 1/k').
        Default is 'without 1/k'.

    Returns
    -------
    float
        The calculated energy of the field.

    Notes
    -----
    The energy is calculated using the formula:

    W = 0.5 * epsilon_0 * c**2 * dVk / (2 * pi)**3 * sum(abs(a1)**2 + abs(a2)**2)
    
    If mode is 'with 1/k', the formula is:
    
    W = 0.5 * epsilon_0 * c**2 * dVk / (2 * pi)**3 * sum(k**2 * {abs(a1)**2 + 
    abs(a2)**2})
    """
    a = "(a1.real**2 + a1.imag**2 + a2.real**2 + a2.imag**2)"
    expr = f"sum({a})" if mode == "without 1/k" else f"sum(k**2 * {a})"
    W = 0.5 * epsilon_0 * c**2 * dVk / (2 * pi) ** 3 * ne.evaluate(expr)
    return W


def convert_tau(tau, mode="1/e^2"):
    """
    Convert between different pulse duration conventions.

    Parameters
    ----------
    tau : float
        Pulse duration.
    mode : str, optional
        The convention of the input pulse duration. Options are '1/e^2',
        'FWHM', 'FWHM-Intensity', and 'std'. Default is '1/e^2'.

    Returns
    -------
    dict
        Dictionary containing the pulse duration in different conventions:
            - '1/e^2' : float
                Pulse duration in 1/e^2 convention.
            - 'FWHM' : float
                Pulse duration in Full Width at Half Maximum (FWHM) convention.
            - 'FWHM-Intensity' : float
                Pulse duration in FWHM-Intensity convention.
            - 'std' : float
                Pulse duration in standard deviation convention.

    Notes
    -----
    The temporal envelope is assumed to have the form exp(-t**2/(tau/2)**2).
    """
    match mode:
        case "1/e^2":
            result = {'1/e^2': tau, 'FWHM': tau*np.sqrt(np.log(2)),
                      'FWHM-Intensity': tau*np.sqrt(np.log(2)/2),
                      'std': tau/(2*np.sqrt(2))}
        case "FWHM":
            result = {'1/e^2': tau/np.sqrt(np.log(2)), 'FWHM': tau,
                      'FWHM-Intensity': tau/np.sqrt(2),
                      'std': tau/(2*np.sqrt(2*np.log(2)))}
        case "FWHM-Intensity":
            result = {'1/e^2': tau/np.sqrt(np.log(2)/2), 'FWHM': tau*np.sqrt(2),
                      'FWHM-Intensity': tau, 'std': tau/(2*np.sqrt(np.log(2)))}
        case "std":
            result = {'1/e^2': 2*np.sqrt(2)*tau, 'FWHM': 2*tau*np.sqrt(2*np.log(2)),
                      'FWHM-Intensity': 2*tau*np.sqrt(np.log(2)), 'std': tau}
        case _:
            raise NotImplementedError(f"Mode {mode} is not supported.")
    result = {k: float(v) for k, v in result.items()}
    return result


def get_max_edge_amplitude(E):
    """
    Calculate the maximum field amplitude at the edges of the grid.

    Parameters
    ----------
    E : tuple of np.array
        Electric field components as a tuple of three numpy arrays (Ex, Ey, Ez).

    Returns
    -------
    float
        The maximum field amplitude at the edges of the grid.

    Notes
    -----
    Electric field is assumed to be complex.
    """
    Ex, Ey, Ez = E
    nx, ny, nz = Ex.shape
    amplitude = np.sqrt(abs(Ex)**2 + abs(Ey)**2 + abs(Ez)**2)
    edges = [np.max(amplitude[0,:,:]), np.max(amplitude[-1,:,:]),
             np.max(amplitude[:,0,:]), np.max(amplitude[:,-1,:]),
             np.max(amplitude[:,:,0]), np.max(amplitude[:,:,-1])]
    amplitude_edge = np.max(edges)
    print(f"Max intensity at edges: {amplitude_edge/amplitude.max():.2e}")
    return amplitude_edge


def get_intensity(field, t):
    """
    Calculate the field intensity at a given time.

    Parameters
    ----------
    field : object
        Field object with a ``calculate_field`` method.
    t : float
        The time at which to calculate the field intensity.

    Returns
    -------
    np.array
        The calculated field intensity.

    Notes
    -----
    The intensity is calculated as (abs(E)**2 + abs(B)**2)/2
    """
    E, B = field.calculate_field(t=t)
    E, B = [np.real(Ex) for Ex in E], [np.real(Bx) for Bx in B]
    intensity = (E[0]**2 + E[1]**2 + E[2]**2 + B[0]**2 + B[1]**2 + B[2]**2)/2
    return intensity


"""
Calculation of analytic scalings from articles.

Currently supports:

1. Total signal and perpendicular signal for two colliding 
paraxial gaussian beams (off-axis).

2. More exact scaling for on-axis collision of two beams.

----

.. [1] F. Karbstein, et al. "Vacuum birefringence at x-ray free-electron 
    lasers." NJP 23.9 (2021): 095001.

.. [2] E. Mosman, F.Karbstein "Vacuum birefringence and diffraction at 
    XFEL: from analytical estimates to optimal parameters." PRD 104.1 
    (2021): 013006.
"""

import numpy as np
from scipy.constants import alpha, c, hbar, m_e, pi
from scipy.integrate import quad
from scipy.special import erfc

W_e = m_e * c**2  # electron's rest energy
lam_C = hbar / (m_e * c)  # reduced Compton wavelength


def get_two_paraxial_scaling(fields, channels="both"):
    """
    Calculate the total signal and perpendicular signal for two colliding 
    paraxial Gaussian beams (off-axis).

    Parameters
    ----------
    fields : list of dict
        List of dictionaries containing the field parameters.
    channels : str, optional
        Channels to iterate through, by default "both".

    Returns
    -------
    N_signal : float
        Total signal.
    N_perp : float
        Perpendicular signal.

    Notes
    -----
    We introduced additional polarization-dependent factors (beta). The 
    formula should work for theta not close to 180, loose focusing and 
    short pulse duration. Result for N_perp should be used only for beta=45.

    This is a variation of Eq.(25) from [1]_.
    """
    theta_c = (fields[1]["theta"] - fields[0]["theta"]) * pi / 180
    beta = (fields[1]["beta"] - fields[0]["beta"]) * pi / 180

    omega = 2 * pi * c / fields[0]["lam"]
    N_signal, N_perp = 0, 0
    # Iterate through pump/probe channels
    match channels:
        case "first":
            fields_to_iterate = [fields]
        case "second":
            fields_to_iterate = [fields[::-1]]
        case _:
            fields_to_iterate = [fields, fields[::-1]]

    for field1, field2 in fields_to_iterate:
        tau1, tau2 = field1["tau"], field2["tau"]
        w01, w02 = field1["w0"], field2["w0"]
        W1, W2 = field1["W"], field2["W"]

        # Define additional parameters
        T1 = tau1 * np.sqrt(1 + 2 * (tau1 / tau2) ** 2)
        T2 = tau2 * np.sqrt(1 + 0.5 * (tau2 / tau1) ** 2)
        w1 = w01 * np.sqrt(1 + 2 * (w01 / w02) ** 2)
        w2 = w02 * np.sqrt(1 + 0.5 * (w02 / w01) ** 2)
        sin_term = T1 * T2 * c**2 / (w1 * w2) * np.sin(theta_c) ** 2
        cos_term = 4 * (1 - np.cos(theta_c)) ** 2

        # Calculate N_signal and N_perp
        theta_term = (1 - np.cos(theta_c)) ** 4 / np.sqrt(cos_term + sin_term)
        theta_term *= 1 / np.sqrt((T1 / tau1 * w01 / w1) ** 2 * cos_term + sin_term)
        dims = hbar * omega / W_e
        prefactor = 8 * pi**2 / 225 * (alpha / pi) ** 4 * W1 * W2**2 / W_e**3 * dims
        prefactor *= T1 / tau1 * w1 / w01 * lam_C**4 / (w1 * w2) ** 2 / 9
        N_perp += prefactor * theta_term * 9 * np.sin(2 * beta) ** 2
        N_signal += prefactor * theta_term * (130 - 66 * np.cos(2 * beta))
    return N_signal, N_perp


def _onaxis_function(x, r, k0=20):
    """
    Calculate the integral for the function defined in [2]_.

    Parameters
    ----------
    x : float
        Parameter x.
    r : float
        Parameter r.
    k0 : float, optional
        Integration limit for the integral, by default 20.

    Returns
    -------
    result : float
        Result of the integral.
    """
    def integrand(k):
        s = 0
        for n in [-1, 1]:
            s += np.exp(2*n*r*x*k) * erfc(n*r*k + x)
        s = np.abs(s)**2
        return np.exp(-k**2) * s
        
    prefactor = np.sqrt((1 + 2*r**2)/3) * x**2 * np.exp(2*x**2)
    result = quad(integrand, -k0, k0)
    return prefactor * result[0]


def get_onaxis_scaling(fields, k0=20):
    """
    Calculate the more exact scaling for on-axis collision of two beams.

    Parameters
    ----------
    fields : list of dict
        List of dictionaries containing the field parameters.
    k0 : float, optional
        Integration limit for the integral in F function, by default 20.

    Returns
    -------
    N_signal : float
        Total signal.
    N_perp : float
        Perpendicular signal.

    Notes
    -----
    We assume that the 1st field in the list is x-ray.

    This is an integrated version of Eq.(11) [Eq. (13)] from [2]_.
    """
    wx, w0 = fields[0]["w0"], fields[1]["w0"]
    beta = wx / w0
    Wx, W = fields[0]["W"], fields[1]["W"]
    lam_x, lam = fields[0]["lam"], fields[1]["lam"]
    omega_x = 2 * pi * c / lam_x
    Nx = Wx / (hbar * omega_x)
    T, tau = fields[0]["tau"], fields[1]["tau"]
    zR = pi * w0**2 / lam

    x0 = 4 * zR / (c * np.sqrt(T**2 + 1/2*tau**2))
    r0 = T / tau
    Fbeta = _onaxis_function(x0 * np.sqrt(1+2*beta**2), r0, k0=k0)
    F0 = _onaxis_function(x0, r0, k0=k0)

    prefactor = 4 * alpha**4 / (3*pi)**1.5 * W**2 * (hbar * omega_x)**2 / W_e**4
    result = prefactor * (lam_C / w0)**4 / (1 + 2*beta**2) * np.sqrt(Fbeta*F0) * Nx

    N_signal = 26/45 * result
    N_perp = 1/25 * result
    return N_signal, N_perp



#!/usr/bin/env python3
"""
Script to create images of participating fields
"""
import argparse
from copy import deepcopy
import gc
import os
from pathlib import Path

from quvac.field.external_field import ExternalField
from quvac.grid import setup_grids
from quvac.plotting import plot_fields
from quvac.simulation import set_precision
from quvac.utils import read_yaml


def parse_args():
    description = "Plot fields present in the simulation"
    argparser = argparse.ArgumentParser(description=description)
    argparser.add_argument(
        "--input", "-i", default=None, help="Input yaml file with field and grid params"
    )
    argparser.add_argument(
        "--output", "-o", default=None, help="Path to save simulation data to"
    )
    return argparser.parse_args()


def image_fields(ini_file, save_path=None):
    if save_path is None:
        save_path = os.path.join(os.path.dirname(ini_file), 'imgs')
    Path(save_path).mkdir(parents=True, exist_ok=True)

    ini_config = read_yaml(ini_file)
    fields_params = ini_config['fields']
    grid_params = ini_config["grid"]

    perf_params = ini_config.get("performance", {})
    precision = perf_params.get("precision", "float32")
    set_precision(precision)

    # Setup grids
    grid_xyz, grid_t = setup_grids(deepcopy(fields_params), deepcopy(grid_params))
    grid_xyz.get_k_grid()
    grid_shape = grid_xyz.grid_shape
    print(f"Grid shape: {grid_shape}")
    print(f"Time steps: {len(grid_t)}")

    plot_keys = ["Intensity"]
    # Setup fields and plot
    for idx,field_param in enumerate(fields_params.values()):
        field = ExternalField([field_param], grid_xyz)
        for t0 in [0, grid_t[-1]]:
            save_loc = os.path.join(save_path, f"field_{idx+1}_t0_{t0*1e15:.1f}fs")
            plot_fields(field, t=t0, plot_keys=plot_keys, norm_lim=1e-10,
                        save_path=save_loc)
        del field
        gc.collect()
    print("Plotting finished successfully")
    

if __name__ == "__main__":
    args = parse_args()
    image_fields(args.input, args.output)


"""
Common interfaces for external field and pump-probe fields to unify
various field types.
"""

import logging
import os

from quvac.field import ANALYTIC_FIELDS
from quvac.field.abc import Field
from quvac.field.maxwell import MaxwellMultiple

_logger = logging.getLogger("simulation")


class ExternalField(Field):
    """
    Class to unite several participating external fields under one interface.

    Parameters
    ----------
    fields_params : list of dict
        List of dictionaries containing the parameters for each external field.
    grid : quvac.grid.GridXYZ
        Spatial and spectral grid.
    nthreads : int, optional
        Number of threads to use for calculations. If not provided, defaults to the 
        number of CPU cores.

    Attributes
    ----------
    fields : list
        List of field objects.
    grid_xyz : tuple of np.array
        The spatial grid.
    nthreads : int
        Number of threads to use for calculations.
    """

    def __init__(self, fields_params, grid, nthreads=None):
        self.fields = []
        self.grid_xyz = grid

        self.nthreads = nthreads if nthreads else os.cpu_count()

        maxwell_params = [
            params
            for params in fields_params
            if params["field_type"].endswith("maxwell")
        ]
        new_params = [
            params
            for params in fields_params
            if not params["field_type"].endswith("maxwell")
        ]
        if maxwell_params:
            new_params.append(maxwell_params)

        _logger.info(
            f"{self.__class__.__name__}\n"
            "----------------------------------------------------"
        )
        for field_params in new_params:
            self.setup_field(field_params)
        _logger.info("----------------------------------------------------")

    def setup_field(self, field_params):
        """
        Set up a field based on the provided parameters.

        Parameters
        ----------
        field_params : dict or list of dict
            Dictionary or list of dictionaries containing the parameters for the field.

        Raises
        ------
        NotImplementedError
            If the field type is not supported.
        """
        if isinstance(field_params, list):
            field_type = "maxwell"
        else:
            field_type = field_params["field_type"]

        if field_type in ANALYTIC_FIELDS:
            field = ANALYTIC_FIELDS[field_type](field_params, self.grid_xyz)
        elif field_type == "maxwell":
            field = MaxwellMultiple(
                    field_params, self.grid_xyz, nthreads=self.nthreads
                )
        else:
            raise NotImplementedError(
                    f"We do not support '{field_type}' field type"
                )

        self.fields.append(field)
        _logger.info(f"Base class: {field.__class__.__name__}")

    def calculate_field(self, t, E_out=None, B_out=None):
        """
        Calculates the electric and magnetic fields at a given time step.
        """
        for field in self.fields:
            E_out, B_out = field.calculate_field(t, E_out=E_out, B_out=B_out)
        return E_out, B_out


class ProbePumpField(Field):
    """
    Class for splitting fields into probe and pump.

    Parameters
    ----------
    fields_params : list of dict
        List of dictionaries containing the parameters for each external field.
    grid : quvac.grid.GridXYZ
        Spatial and spectral grid.

    probe_pump_idx : dict, optional
        Dictionary specifying which fields are probe and pump. Required keys are:
            - 'probe' : list of int
                Indices of the fields to be used as probe.
            - 'pump' : list of int
                Indices of the fields to be used as pump.
                
        If not provided, defaults to {"probe": [0], "pump": [1]}.

    nthreads : int, optional
        Number of threads to use for calculations. If not provided, defaults to the 
        number of CPU cores.

    Attributes
    ----------
    probe_pump_idx : dict
        Dictionary specifying which fields are probe and pump.
    nthreads : int
        Number of threads to use for calculations.
    probe_field : ExternalField
        ExternalField object for the probe fields.
    pump_field : ExternalField
        ExternalField object for the pump fields.
    """

    def __init__(self, fields_params, grid, probe_pump_idx=None, nthreads=None):
        if not probe_pump_idx:
            probe_pump_idx = {"probe": [0], "pump": [1]}
        self.probe_pump_idx = probe_pump_idx

        self.nthreads = nthreads if nthreads else os.cpu_count()

        probe_params = [fields_params[idx] for idx in probe_pump_idx["probe"]]
        pump_params = [fields_params[idx] for idx in probe_pump_idx["pump"]]

        self.probe_field = ExternalField(probe_params, grid, nthreads=nthreads)
        self.pump_field = ExternalField(pump_params, grid, nthreads=nthreads)

    def calculate_field(self, t, E_probe=None, B_probe=None, E_pump=None, B_pump=None):
        """
        Calculate the probe and pump fields at a given time t.

        Parameters
        ----------
        t : float
            The time at which to calculate the fields.
        E_probe : np.array, optional
            Array to store the electric field output for the probe.
        B_probe : np.array, optional
            Array to store the magnetic field output for the probe.
        E_pump : np.array, optional
            Array to store the electric field output for the pump.
        B_pump : np.array, optional
            Array to store the magnetic field output for the pump.

        Returns
        -------
        tuple of tuple of np.array
            The calculated electric and magnetic fields for the probe and pump.
        """
        self.probe_field.calculate_field(t, E_out=E_probe, B_out=B_probe)
        self.pump_field.calculate_field(t, E_out=E_pump, B_out=B_pump)
        return (E_probe, B_probe), (E_pump, B_pump)



"""
Analyzer class that calculate different observables from amplitudes.

Currently supports:

1. Total (polarization insensitive) signal.
2. Polarization sensitive signal:
    - As a projection to fixed optical axis.
    - As a projection to local polarization axis.
    - 3 stokes parameters (P1, P2, P3) for arbitrary 
      linear polarization basis.

3. Discernible signal.
4. Background field spectra.
"""
# ruff: noqa: F841
import logging
import os
import warnings

import numexpr as ne
import numpy as np
from scipy.constants import c, epsilon_0, hbar, pi
from scipy.integrate import trapezoid
from scipy.ndimage import map_coordinates

from quvac import config
from quvac.field.external_field import ExternalField
from quvac.field.maxwell import MaxwellMultiple
from quvac.grid import GridXYZ, get_pol_basis, setup_grids
from quvac.log import sph_interp_warn
from quvac.utils import read_yaml

_logger = logging.getLogger("simulation")


def get_polarization_vector(theta, phi, beta):
    """
    Calculate the polarization vector.

    Parameters
    ----------
    theta : float
        Polar angle (in radians).
    phi : float
        Azimuthal angle (in radians).
    beta : float
        Polarization angle (in radians).

    Returns
    -------
    ep : numpy.ndarray
        Polarization vector.
    """
    e1, e2 = get_pol_basis(theta, phi)
    ep =(e1 * np.cos(beta, dtype=config.FDTYPE) + 
         e2 * np.sin(beta, dtype=config.FDTYPE))
    return ep


def sph2cart(r, theta, phi):
    """
    Convert spherical coordinates to Cartesian coordinates.

    Parameters
    ----------
    r : numpy.ndarray
        Radial distance.
    theta : numpy.ndarray
        Polar angle.
    phi : numpy.ndarray
        Azimuthal angle.

    Returns
    -------
    x, y, z : numpy.ndarray
        Cartesian coordinates.
    """
    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta) * np.ones_like(phi)
    return x, y, z


def xyz2idx(xyz, xyz_grid):
    """
    Convert Cartesian coordinates to array indices.

    Parameters
    ----------
    xyz : tuple of numpy.ndarray
        Cartesian coordinates.
    xyz_grid : tuple of numpy.ndarray
        Cartesian grid.

    Returns
    -------
    idxs : numpy.ndarray
        Array indices.
    """
    nx, ny, nz = xyz[0].shape
    idxs = np.empty((3, nx, ny, nz))
    for i, (x, grid) in enumerate(zip(xyz, xyz_grid, strict=True)):
        x0, x1 = grid[0], grid[-1]
        idxs[i] = (x - x0) / (x1 - x0) * (len(grid) - 1)
    return idxs


def cartesian_to_spherical_array(
    arr, xyz_grid, spherical_grid=None, angular_resolution=None,
    **interp_kwargs
):
    """
    Transform an array with data on Cartesian grid to the array with data 
    on spherical grid.

    Parameters
    ----------
    arr : numpy.ndarray
        Array with data on Cartesian grid.
    xyz_grid : quvac.grid.GridXYZ
        Cartesian grid object.
    spherical_grid : tuple or str, optional
        Spherical grid or path to the file containing the spherical grid, 
        by default None.
    angular_resolution : float, optional
        Angular resolution, by default None.
    **interp_kwargs : dict
        Additional interpolation parameters. Currently not used, implement 
        in the future.

    Returns
    -------
    spherical_grid : tuple
        Spherical grid.
    arr_sph : numpy.ndarray
        Array with data on spherical grid.
    """
    # Calculate spherical grid if not given
    if not spherical_grid:
        dk = np.min(xyz_grid.dkxkykz)
        kmax = np.max(xyz_grid.kabs)
        dangle = angular_resolution if angular_resolution else 1.0 * pi / 180

        k = np.arange(0.0, kmax, dk, dtype=config.FDTYPE)
        theta = np.arange(0.0, pi, dangle, dtype=config.FDTYPE)
        phi = np.arange(0.0, 2 * pi, dangle, dtype=config.FDTYPE)
        spherical_grid = (k, theta, phi)
    elif isinstance(spherical_grid, str) and os.path.isfile(spherical_grid):
        data = np.load(spherical_grid)
        spherical_grid = (data["k"], data["theta"], data["phi"])
    spherical_mesh = np.meshgrid(*spherical_grid, indexing="ij", sparse=True)

    # Find corresponding cartesian coordinates of spherical mesh:
    # (r,theta,phi) -> (x, y, z)
    xyz_for_sph = sph2cart(*spherical_mesh)

    # Convert cartesian coordinates to array idx
    idxs = xyz2idx(xyz_for_sph, xyz_grid.kgrid_shifted)

    # Interpolate data on a desired grid
    # interpolation_kwargs should be implemented here
    arr_sph = map_coordinates(arr, idxs, order=1)
    return spherical_grid, arr_sph


def integrate_spherical(arr, axs, axs_names=("k", "theta", "phi"),
                        axs_integrate=("k", "theta", "phi")):
    """
    Integrate an array over spherical coordinates.

    Parameters
    ----------
    arr : numpy.ndarray
        Array to be integrated.
    axs : list of numpy.ndarray
        List of axes.
    axs_names : list of str, optional
        List of axis names, by default ["k", "theta", "phi"].
    axs_integrate : list of str, optional
        List of axes to integrate over, by default ["k", "theta", "phi"].

    Returns
    -------
    integrand : numpy.ndarray
        Integrated array.

    Raises
    ------
    AssertionError
        Axes of array and axs names do not match.

    Notes
    -----
    We assume that the array is given on a spherical grid without
    the jacobian factor, therefore the array is mutiplied by k**2 sin(theta).
    
    The function uses trapezoidal rule for integration.
    """
    err_msg = "Axes of array and axs names do not match"
    condition = len(axs) == len(axs_names) and len(axs) >= len(axs_integrate)
    assert condition, err_msg

    axs_names_ = list(axs_names)

    integrand = arr.copy()
    for ax_name in axs_integrate:
        idx = axs_names.index(ax_name)
        idx_ = axs_names_.index(ax_name)

        ax_shape = [1 for _ in range(len(axs_names_))]
        ax_shape[idx_] = integrand.shape[idx_]

        ax = axs[idx].reshape(ax_shape)
        if ax_name == "k":
            integrand *= ax**2
        elif ax_name == "theta":
            integrand *= np.sin(ax)
        integrand = trapezoid(integrand, axs[idx], axis=idx_)
        axs_names_.pop(idx_)
    return integrand


def _get_detector_idx(phi, theta, phi0, theta0, dphi, dtheta):
    # consider detector regions that lie on the line phi=0 or phi=2*pi
    if phi0-dphi < 0:
        idx_phi = (phi <= phi0+dphi) + (phi >= 2*pi-abs(phi0-dphi))
    elif phi0+dphi > 2*pi:
        idx_phi = (phi >= phi0-dphi) + (phi <= phi0+dphi-2*pi)
    else:
        idx_phi = (phi >= phi0-dphi) * (phi <= phi0+dphi)
    
    idx_theta = (theta >= theta0-dtheta) * (theta <= theta0+dtheta)
    return idx_phi, idx_theta


def signal_in_detector(dN, theta, phi, detector, align_to_max=False):
    """
    Calculate the signal detected within a specified detector region.

    Parameters
    ----------
    dN : numpy.ndarray
        Differential angular photon spectrum on a spherical grid.
    theta : numpy.ndarray
        Array of polar angles (in radians) corresponding to the spherical grid.
    phi : numpy.ndarray
        Array of azimuthal angles (in radians) corresponding to the spherical grid.
    detector : dict
        Dictionary specifying the detector region. Required keys are:
            - 'phi0' : float
                Central azimuthal angle of the detector (in degrees).
            - 'theta0' : float
                Central polar angle of the detector (in degrees).
            - 'dphi' : float
                Half-width of the azimuthal angle range (in degrees).
            - 'dtheta' : float
                Half-width of the polar angle range (in degrees).
    align_to_max : bool
        Whether to align detector to the max spot in the detected region, 
        by default False.

    Returns
    -------
    float
        The total signal detected within the specified detector region.

    Notes
    -----
    The function integrates the signal over the specified detector region in spherical 
    coordinates.
    """
    phi0, theta0, dphi0, dtheta0 = [np.radians(detector[key]) for key 
                                  in "phi0 theta0 dphi dtheta".split()]
    idx_phi,idx_theta = _get_detector_idx(phi, theta, phi0, theta0, dphi0, dtheta0)
    dphi, dtheta = [ax[1]-ax[0] for ax in (phi, theta)]
    
    dN_det = dN[idx_theta][:,idx_phi]
    theta_det, phi_det = theta[idx_theta], phi[idx_phi]

    if align_to_max:
        max_index = np.argmax(dN_det)
        theta_max, phi_max = np.unravel_index(max_index, dN_det.shape)
        phi0_new, theta0_new = phi_det[phi_max], theta_det[theta_max]
        idx_phi,idx_theta = _get_detector_idx(phi, theta, phi0_new, theta0_new,
                                              dphi0, dtheta0)
        dN_det = dN[idx_theta][:,idx_phi]
        theta_det, phi_det = theta[idx_theta], phi[idx_phi]

    N_detected = np.sum(dN_det * np.sin(theta_det)[:,None]) * dphi * dtheta
    return N_detected
    

def get_simulation_fields(ini_file):
    """
    Get simulation fields from an initialization file.

    Parameters
    ----------
    ini_file : str
        Path to the initialization file.

    Returns
    -------
    fields : list of quvac.field.ExternalField
        List of field objects.
    """
    ini = read_yaml(ini_file)
    fields_params = ini["fields"]
    grid_params = ini["grid"]

    grid_xyz, grid_t = setup_grids(fields_params, grid_params)
    grid_xyz.get_k_grid()

    fields = []
    for field_params in fields_params.values():
        field = ExternalField([field_params], grid_xyz)
        fields.append(field)
    return fields


def transform_S12_to_a12(S1, S2, k, transform="forward"):
    """
    Calculate spectral coefficients a1 and a2 from complex signal
    amplitudes S1 and S2.

    Parameters
    ----------
    S1 : numpy.ndarray
        Complex signal amplitude S1.
    S2 : numpy.ndarray
        Complex signal amplitude S2.
    k : numpy.ndarray
        Wave number.
    transform : str, optional
        Transformation direction, by default "forward".

    Returns
    -------
    a1, a2 : numpy.ndarray
        Spectral coefficients a1 and a2.

    Notes
    -----
    The backward transform corresponds to a12 -> S12.
    """
    prefactor = np.sqrt(2*hbar*k/(epsilon_0*c))
    if transform == "backward":
        prefactor = np.where(prefactor != 0, 1 / prefactor, 0.)
    a1 = S1 * prefactor
    a2 = S2 * prefactor
    return a1, a2


def get_spectra_from_Stokes(P0, P1, P2, P3, basis="linear"):
    """
    Calculate spectra from Stokes parameters.

    Parameters
    ----------
    P0 : numpy.ndarray
        Stokes parameter P0.
    P1 : numpy.ndarray
        Stokes parameter P1.
    P2 : numpy.ndarray
        Stokes parameter P2.
    P3 : numpy.ndarray
        Stokes parameter P3.
    basis : str, optional
        Polarization basis, by default "linear".

    Returns
    -------
    Nf, Np : numpy.ndarray
        Parallel and perpendicular-polarized spectra.
    """
    match basis:
        case "linear":
            P = P1
        case "linear-45":
            P = P2
        case "circular":
            P = P3
    Nf = 0.5 * (P0 + P)
    Np = 0.5 * (P0 - P)
    return Nf, Np


class VacuumEmissionAnalyzer:
    """
    Calculates spectra and observables from amplitudes.

    Amplitudes are provided by 
    ``quvac.integrator.vacuum_emission.VacuumEmission`` class.

    Parameters
    ----------
    fields_params : dict
        Dictionary containing the field parameters.
    data_path : str
        Path to the data file.
    save_path : str, optional
        Path to save the results, by default None.

    Attributes
    ----------
    fields_params : dict
        Dictionary containing the field parameters.
    grid_xyz : quvac.grid.GridXYZ
        The spatial grid object.
    S1 : numpy.ndarray
        Complex signal amplitude S1.
    S2 : numpy.ndarray
        Complex signal amplitude S2.
    save_path : str
        Path to save the results.
    """

    def __init__(self, fields_params, data_path, save_path=None):
        self.fields_params = fields_params
        # Load data
        data = np.load(data_path)
        grid = (data["x"], data["y"], data["z"])
        self.grid_xyz = GridXYZ(grid)
        self.grid_xyz.get_k_grid()
        # Update local dict with variables from GridXYZ class
        self.__dict__.update(self.grid_xyz.__dict__)

        for ax in "xyz":
            kx = getattr(self, f"k{ax}")
            setattr(self, f"k{ax}", np.fft.fftshift(kx))
            # self.__dict__[f"k{ax}"] = np.fft.fftshift(kx)

        self.S1, self.S2 = (data["S1"].astype(config.CDTYPE),
                            data["S2"].astype(config.CDTYPE))

        self.save_path = save_path

    def get_total_signal(self):
        """
        Calculates the total signal.
        """
        S = self.S1.real**2 + self.S1.imag**2 + self.S2.real**2 + self.S2.imag**2
        self.N_xyz = np.fft.fftshift(S / (2 * pi) ** 3)

        self.N_total = np.sum(self.N_xyz) * self.dVk

    def get_polarization_from_field(self):
        """
        Calculates the polarization from the field.

        Returns
        -------
        tuple
            Tuple containing the polarization components (efx, efy, efz).
        """
        perp_field_params = self.fields_params[self.perp_field_idx]
        field = MaxwellMultiple([perp_field_params], self.grid_xyz)
        a1, a2 = field.a1, field.a2
        Ex = ne.evaluate("e1x*a1 + e2x*a2", global_dict=self.__dict__)
        Ey = ne.evaluate("e1y*a1 + e2y*a2", global_dict=self.__dict__)
        Ez = ne.evaluate("e1z*a1 + e2z*a2", global_dict=self.__dict__)
        E = ne.evaluate("sqrt(Ex**2 + Ey**2 + Ez**2)")
        E_inv = ne.evaluate("where((Ex==0) & (Ey==0) & (Ez==0), 0, 1/E)")
        # E_inv = np.nan_to_num(1. / E)
        efx, efy, efz = Ex * E_inv, Ey * E_inv, Ez * E_inv
        return (efx, efy, efz)

    def _get_polarization_vector(self, angles, perp_type="optical axis"):
        """
        Calculates polarization vectors parallel to chosen axis and 
        perpendicular to it.

        Parameters
        ----------
        angles : tuple
            Tuple containing the angles (theta, phi, beta).
        perp_type : str, optional
            Type of perpendicular polarization, by default "optical axis".

        Returns
        -------
        tuple
            Tuple containing the parallel and perpendicular polarization 
            components.
        """
        if perp_type == "optical axis":
            self.efx, self.efy, self.efz = get_polarization_vector(*angles)
            angles[-1] += pi / 2
            self.epx, self.epy, self.epz = get_polarization_vector(*angles)
        elif perp_type == "local axis":
            efx, efy, efz = self.get_polarization_from_field()
            self.efx, self.efy, self.efz = efx, efy, efz
            kx, ky, kz = [ne.evaluate("where(kabs==0, 0, k/kabs)", 
                                      global_dict={"kabs": self.kabs, "k": k})
                          for k in self.kmeshgrid]
            self.epx = ne.evaluate("ky*efz - kz*efy")
            self.epy = ne.evaluate("kz*efx - kx*efz")
            self.epz = ne.evaluate("kx*efy - ky*efx")
        else:
            raise NotImplementedError(f"{perp_type} is not implemented")
        return (self.efx, self.efy, self.efz), (self.epx, self.epy, self.epz)

    def get_perp_signal(self, angles, perp_type="optical axis",
                        stokes=False):
        """
        Calculates the perpendicular signal.

        Parameters
        ----------
        angles : tuple
            Tuple containing the angles (theta, phi, beta) in degrees.
        perp_type : str, optional
            Type of perpendicular polarization, by default "optical axis".
        stokes : bool, optional
            Whether to calculate Stokes parameters, by default False.
        """
        angles = [angle * pi / 180 for angle in angles]

        # get one polarization direction to project on:
        self.ef, self.ep = self._get_polarization_vector(angles, perp_type)
        epx, epy, epz = self.ep
        e1x, e1y, e1z = self.e1x, self.e1y, self.e1z
        e2x, e2y, e2z = self.e2x, self.e2y, self.e2z

        # Calculate perp signal or stokes parameters
        if stokes:
            self.get_Stokes_vector_for_ep()
        else:
            Sp = ((epx*e1x + epy*e1y + epz*e1z)*self.S1 + 
                  (epx*e2x + epy*e2y + epz*e2z)*self.S2)
            Sp = Sp.real**2 + Sp.imag**2

            self.Np_xyz = np.fft.fftshift(Sp / (2 * pi) ** 3)
            self.Np_total = np.sum(self.Np_xyz) * self.dVk

    def get_Stokes_vector_for_ep(self):
        '''
        Given amplitudes S1,S2 for arbitrary linear polarization
        basis e1,e2, calculate Stokes vectors for detector ep.

        For given ef (||) and ep (perp):
            - P0 = N_xyz = (S_||)**2 + (S_perp)**2

            - P1 = (S_||)**2 - (S_perp)**2
            
            - P2 = (S_45)**2 - (S_135)**2     - difference of linear 
              polarizations in 45 basis.
            
            - P3 = (S_right)**2 - (S_left)**2 - difference of circular 
              polarizations.

        '''
        e1x, e1y, e1z = self.e1x, self.e1y, self.e1z
        e2x, e2y, e2z = self.e2x, self.e2y, self.e2z
        efx, efy, efz = self.ef
        epx, epy, epz = self.ep

        Sf = ((efx*e1x + efy*e1y + efz*e1z)*self.S1 + 
              (efx*e2x + efy*e2y + efz*e2z)*self.S2)
        Sp = ((epx*e1x + epy*e1y + epz*e1z)*self.S1 + 
              (epx*e2x + epy*e2y + epz*e2z)*self.S2)

        P1 = Sf.real**2 + Sf.imag**2 - (Sp.real**2 + Sp.imag**2)
        P2 = 2 * np.real(Sf * np.conj(Sp))
        P3 = -2 * np.imag(Sf * np.conj(Sp))

        self.P1, self.P2, self.P3 = [np.fft.fftshift(P / (2 * pi) ** 3)
                                     for P in (P1,P2,P3)]

    def get_signal_on_sph_grid(
        self, key="N_xyz", spherical_grid=None, angular_resolution=None,
        check_total=False, **interp_kwargs
    ):
        """
        Transforms an array with data on Cartesian grid to the array with 
        data on spherical grid.

        Parameters
        ----------
        key : str, optional
            Key for the data array, by default "N_xyz".
        spherical_grid : tuple or str, optional
            Spherical grid or path to the file containing the spherical grid,
            by default None.
        angular_resolution : float, optional
            Angular resolution, by default None.
        check_total : bool, optional
            Whether to check the total signal, by default False.
        **interp_kwargs : dict
            Additional interpolation parameters.
        """
        arr = getattr(self, key)
        spherical_grid, N_sph = cartesian_to_spherical_array(
            arr,
            self.grid_xyz,
            spherical_grid=spherical_grid,
            angular_resolution=angular_resolution,
            **interp_kwargs,
        )
        if "xyz" in key:
            sph_key = key.replace("xyz", "sph")
            total_key = key.replace("xyz", "total")
        else:
            sph_key = key + '_sph'
            total_key = key + "_total"
        sph_total_key = f"{sph_key}_total"
        setattr(self, sph_key, N_sph)

        if check_total:
            N_total = integrate_spherical(N_sph, spherical_grid)
            setattr(self, sph_total_key, N_total)

            if not np.isclose(getattr(self, total_key), N_total, rtol=1e-2):
                warn_message = sph_interp_warn.format(
                    total_key, getattr(self, total_key), N_total
                )
                warnings.warn(warn_message, stacklevel=2)
                _logger.warning(warn_message)

        self.spherical_grid = self.k, self.theta, self.phi = spherical_grid

    def get_photon_spectrum_from_a12(self, a1, a2):
        """
        Calculates the photon spectrum from spectral coefficients a1 and a2.

        Parameters
        ----------
        a1 : numpy.ndarray
            Spectral coefficient a1.
        a2 : numpy.ndarray
            Spectral coefficient a2.

        Returns
        -------
        numpy.ndarray
            Photon spectrum.
        """
        prefactor = 0.5 * epsilon_0 * c / hbar
        k = self.kabs
        N_xyz = ne.evaluate(
            "prefactor * (a1.real**2 + a1.imag**2 + a2.real**2 + a2.imag**2) / k"
        )
        N_xyz[0, 0, 0] = 0.0
        return np.fft.fftshift(N_xyz)
    
    def get_background_xyz(self, add_to_cls_dict=True, bgr_idx=None):
        """
        Calculates the background field spectra on Cartesian grid.

        Parameters
        ----------
        add_to_cls_dict : bool, optional
            Whether to add the background spectra to the class dictionary, 
            by default True.
        bgr_idx : int, optional
            Index of the background field, by default None.

        Returns
        -------
        numpy.ndarray
            Background field spectra.
        """
        if bgr_idx is not None:
            bgr_field = MaxwellMultiple(self.fields_params[bgr_idx], self.grid_xyz)
        else:
            bgr_field = MaxwellMultiple(self.fields_params, self.grid_xyz)
        bgr_N_xyz = self.get_photon_spectrum_from_a12(bgr_field.a1, bgr_field.a2)
        if add_to_cls_dict:
            self.background_xyz = bgr_N_xyz
        return bgr_N_xyz

    def get_background(self, discernibility="angular", bgr_idx=None,
                       **interp_kwargs):
        """
        Calculates the background field spectra.

        Parameters
        ----------
        discernibility : str, optional
            Type of discernibility, by default "angular".
        bgr_idx : int, optional
            Index of the background field, by default None.
        **interp_kwargs : dict
            Additional interpolation parameters.

        Returns
        -------
        numpy.ndarray
            Background field spectra.
        """
        bgr_N_xyz = self.get_background_xyz(add_to_cls_dict=True, bgr_idx=bgr_idx)

        # Interpolate on spherical grid
        _, bgr_N_sph = cartesian_to_spherical_array(
            bgr_N_xyz,
            self.grid_xyz,
            spherical_grid=self.__dict__.get('spherical_grid', None),
            **interp_kwargs,
        )

        # Integrate over k if needed
        if discernibility == "angular":
            bgr_N_sph = integrate_spherical(
                bgr_N_sph, self.spherical_grid, axs_integrate=["k"]
            )
        return bgr_N_sph

    def get_discernible_signal(self, discernibility="angular"):
        """
        Calculates the discernible signal.

        Parameters
        ----------
        discernibility : str, optional
            Type of discernibility, by default "angular".
        """
        # Calculate numerical background
        self.background = self.get_background(discernibility=discernibility)

        # Integrate signal spectrum if required and determine discernible regions
        if discernibility == "angular":
            self.N_angular = integrate_spherical(
                self.N_sph, self.spherical_grid, axs_integrate=["k"]
            )
            self.discernible = self.N_angular > self.background
        else:
            self.discernible = self.N_sph > self.background

        # Integrate over discernible regions
        self.N_disc = integrate_spherical(
            self.N_sph * self.discernible, self.spherical_grid
        )
    
    def get_signal_a12(self, add_signal_bg=False):
        """
        Calculates the spectral coefficients a1 and a2 from signal amplitudes S1 and S2.

        Parameters
        ----------
        add_signal_bg : bool, optional
            Whether to add the background signal, by default False.
        """
        self.a1_sig, self.a2_sig = transform_S12_to_a12(self.S1, self.S2, self.kabs)
        if add_signal_bg:
            bgr_field = MaxwellMultiple(self.fields_params, self.grid_xyz)
            a1_bgr, a2_bgr = bgr_field.a1, bgr_field.a2
            self.a1_mix, self.a2_mix = self.a1_sig + a1_bgr, self.a2_sig + a2_bgr
            self.a1_mix, self.a2_mix = [np.fft.fftshift(a) for a 
                                        in (self.a1_mix,self.a2_mix)]
        self.a1_sig, self.a2_sig = [np.fft.fftshift(a) for a 
                                    in (self.a1_sig,self.a2_sig)]

    def write_data(self, keys):
        """
        Writes the data to a file.

        Parameters
        ----------
        keys : list of str
            List of keys for the data to be written.
        """
        data = {key: getattr(self, key) for key in keys}
        np.savez(self.save_path, **data)

    def get_total_spectra(
        self,
        calculate_xyz_background=False,
        bgr_idx=None,
        calculate_spherical=False,
        spherical_params=None,
        calculate_discernible=False,
        discernibility="angular",
    ):
        """
        Calculates the total spectra.

        Parameters
        ----------
        calculate_xyz_background : bool, optional
            Whether to calculate the background spectra on Cartesian grid, 
            by default False.
        bgr_idx : int, optional
            Index of the background field, by default None.
        calculate_spherical : bool, optional
            Whether to calculate the spectra on spherical grid,
            by default False.
        spherical_params : dict, optional
            Parameters for the spherical grid, by default None.
        calculate_discernible : bool, optional
            Whether to calculate the discernible signal, by default False.
        discernibility : str, optional
            Type of discernibility, by default "angular".
        """
        self.get_total_signal()
        keys = "kx ky kz N_xyz N_total".split()

        if calculate_spherical:
            self.get_signal_on_sph_grid(key="N_xyz", check_total=True, 
                                        **spherical_params)
            keys.extend("k theta phi N_sph N_sph_total".split())
        if calculate_xyz_background:
            # self.get_background_xyz()
            self.background = self.get_background(discernibility=None, bgr_idx=bgr_idx)
            keys.extend("background_xyz background".split())
        if calculate_discernible:
            self.get_discernible_signal(discernibility)
            keys.extend("background discernible N_disc".split())
        
        self.write_data(keys)

    def get_polarization_spectra(
        self,
        perp_field_idx=1,
        perp_type=None,
        calculate_spherical=False,
        spherical_params=None,
        stokes=False,
    ):
        """
        Calculates the polarization spectra.

        Parameters
        ----------
        perp_field_idx : int, optional
            Index of the perpendicular field, by default 1.
        perp_type : str, optional
            Type of perpendicular polarization, by default None.
        calculate_spherical : bool, optional
            Whether to calculate the spectra on spherical grid,
            by default False.
        spherical_params : dict, optional
            Parameters for the spherical grid, by default None.
        stokes : bool, optional
            Whether to calculate Stokes parameters, by default False.
        """
        self.perp_field_idx = perp_field_idx - 1
        angle_keys = "theta phi beta".split()
        angles = [self.fields_params[self.perp_field_idx][key] for key in angle_keys]
        self.get_perp_signal(angles, perp_type=perp_type, stokes=stokes)
        keys = "kx ky kz epx epy epz efx efy efz".split()

        if not stokes:
            keys.extend("Np_xyz Np_total".split())
            if calculate_spherical:
                self.get_signal_on_sph_grid(key="Np_xyz", check_total=True,
                                             **spherical_params)
                keys.extend("k theta phi Np_sph Np_sph_total".split())
        else:
            stokes_keys = "P1 P2 P3".split()
            keys.extend(stokes_keys)
            if calculate_spherical:
                keys.extend("k theta phi".split())
                for stokes_key in stokes_keys:
                    self.get_signal_on_sph_grid(key=stokes_key, check_total=False,
                                                **spherical_params)
                    keys.extend([f"{stokes_key}_sph"])
        
        self.write_data(keys)

    def get_mix_signal_bg(self, add_signal_bg=False):
        """
        Calculates the mixed signal and background.

        Parameters
        ----------
        add_signal_bg : bool, optional
            Whether to add the background signal, by default False.
        """
        self.get_signal_a12(add_signal_bg)

        keys = "a1_sig a2_sig".split()
        if add_signal_bg:
            keys.extend("a1_mix a2_mix".split())
        self.write_data(keys)

    def get_spectra(
        self,
        mode="total",
        perp_field_idx=1,
        perp_type=None,
        calculate_xyz_background=False,
        bgr_idx=None,
        stokes=False,
        calculate_spherical=False,
        spherical_params=None,
        calculate_discernible=False,
        discernibility="angular",
        add_signal_bg=False,
    ):
        """
        Calculates the spectra based on the specified mode.

        Parameters
        ----------
        mode : str, optional
            Mode of the calculation, by default "total".
        perp_field_idx : int, optional
            Index of the perpendicular field, by default 1.
        perp_type : str, optional
            Type of perpendicular polarization, by default None.
        calculate_xyz_background : bool, optional
            Whether to calculate the background spectra on Cartesian grid,
            by default False.
        bgr_idx : int, optional
            Index of the background field, by default None.
        stokes : bool, optional
            Whether to calculate Stokes parameters, by default False.
        calculate_spherical : bool, optional
            Whether to calculate the spectra on spherical grid,
            by default False.
        spherical_params : dict, optional
            Parameters for the spherical grid, by default None.
        calculate_discernible : bool, optional
            Whether to calculate the discernible signal, by default False.
        discernibility : str, optional
            Type of discernibility, by default "angular".
        add_signal_bg : bool, optional
            Whether to add the background signal, by default False.
        """
        if mode == "total":
            self.get_total_spectra(
                calculate_xyz_background=calculate_xyz_background,
                bgr_idx=bgr_idx,
                calculate_spherical=calculate_spherical,
                spherical_params=spherical_params,
                calculate_discernible=calculate_discernible,
                discernibility=discernibility
            )
        elif mode == "polarization":
            self.get_polarization_spectra(
                perp_field_idx=perp_field_idx,
                perp_type=perp_type,
                calculate_spherical=calculate_spherical,
                spherical_params=spherical_params,
                stokes=stokes,
            )
        elif mode == "mix_signal_bg":
            self.get_mix_signal_bg(add_signal_bg=add_signal_bg)




"""
Abstract interfaces for field classes.

For details about spectral coefficient calculation (a1, a2)
refer to :ref:`implementation` section of documentation. 
"""

from abc import ABC, abstractmethod
import logging

import numexpr as ne
import numpy as np
import pyfftw
from scipy.constants import pi
from scipy.spatial.transform import Rotation

from quvac import config
from quvac.field.utils import get_field_energy, get_field_energy_kspace

ANGLE_KEYS = [
    "theta",
    "phi", 
    "beta", 
    "phase0", 
    "ellipse_rotation_angle",
]
_logger = logging.getLogger("simulation")


class Field(ABC):
    """
    Abstract base class for fields.

    This class defines the interface for calculating fields at a given time step.
    Subclasses must implement the ``calculate_field`` method.

    Parameters
    ----------
    field_params : dict
        Dictionary containing the field parameters.
    grid : quvac.grid.GridXYZ
        Spatial and spectral grid.

    Attributes
    ----------
    grid_xyz : quvac.grid.GridXYZ
        The spatial and spectral grid.
    rotation_m : np.array
        Rotation matrix.
    rotation_bwd_m : np.array
        Inverse rotation matrix.
    """
    def __init__(self, field_params, grid):
        # add grid attributes to field class attributes
        self.grid_xyz = grid
        self.__dict__.update(self.grid_xyz.__dict__)

        # Dynamically create class instance variables available with
        # self.<variable_name>
        for key, val in field_params.items():
            if key in ANGLE_KEYS:
                val *= pi / 180.0
            setattr(self, key, val)
    
    def _check_energy(self):
        """
        Check and adjust the field energy.
        """
        E, B = self.calculate_field(t=0)
        W = get_field_energy(E, B, self.dV)

        self.modify_energy = "W" in self.__dict__.keys() and not np.isclose(W, self.W, 
                                                                            rtol=1e-5)
        if self.modify_energy:
            self.W_correction = np.sqrt(self.W / W) if W > 0 else 0
            self.W_num = self.W
    
    def get_rotation(self):
        """
        Defines the rotation transforming (0,0,1) -> (kx,ky,kz) 
        for vectors and (1,0,0) -> e(beta) = e1*cos(beta) + e2*sin(beta).
        """
        self.beta = getattr(self, "beta", 0)
        self.rotation = Rotation.from_euler("ZYZ", (self.phi, self.theta, self.beta))
        self.rotation_m = self.rotation.as_matrix()
        # Inverse rotation: (kx,ky,kz) -> (0,0,1)
        self.rotation_bwd = self.rotation.inv()
        self.rotation_bwd_m = self.rotation_bwd.as_matrix()

    def rotate_coordinates(self, rotate_grid=True):
        """
        Rotates the coordinate grid.

        Parameters
        ----------
        rotate_grid : bool, optional
            Whether to rotate the grid coordinates. Default is True.
        """
        self.get_rotation()
        if rotate_grid:
            axes = "xyz"
            x_, y_, z_ = self.xyz
            for i, ax in enumerate(axes):
                mx, my, mz = self.rotation_bwd_m[i, :]
                new_ax = ne.evaluate(
                    "mx*(x_-x0) + my*(y_-y0) + mz*(z_-z0)", global_dict=self.__dict__
                )
                setattr(self, ax, new_ax)
        else:
            self.x, self.y, self.z = self.xyz

    def get_EB_out(self, E_out, B_out, mode):
        """
        Get the output arrays for the electric and magnetic fields.

        Parameters
        ----------
        E_out : array-like, optional
            Output array for the electric field.
        B_out : array-like, optional
            Output array for the magnetic field.
        mode : str
            Mode of calculation ('real' or 'complex').

        Returns
        -------
        tuple of array-like
            The output arrays for the electric and magnetic fields.
        """
        dtype = np.float64 if mode == "real" else np.complex128
        if E_out is None:
            E_out = [np.zeros(self.grid_shape, dtype=dtype) for _ in range(3)]
        if B_out is None:
            B_out = [np.zeros(self.grid_shape, dtype=dtype) for _ in range(3)]
        return E_out, B_out
    
    def rotate_fields_back(self, E_out, B_out, mode):
        """
        Rotate the fields back to the original coordinate frame.

        Parameters
        ----------
        E_out : array-like, optional
            Output array for the electric field.
        B_out : array-like, optional
            Output array for the magnetic field.
        mode : str
            Mode of calculation ('real' or 'complex').

        Returns
        -------
        tuple of array-like
            The rotated electric and magnetic fields.
        """
        E_out, B_out = self.get_EB_out(E_out, B_out, mode)

        out_dtype = config.FDTYPE if mode == "real" else config.CDTYPE
        # Transform to the original coordinate frame
        for i, (Ei, Bi) in enumerate(zip(E_out, B_out, strict=2)):
            mx, my, mz = self.rotation_m[i, :]
            Ei += ne.evaluate("mx*Ex + my*Ey + mz*Ez",
                              global_dict=self.__dict__).astype(out_dtype)
            Bi += ne.evaluate("mx*Bx + my*By + mz*Bz",
                              global_dict=self.__dict__).astype(out_dtype)
        return E_out, B_out
    
    def convert_fields_to_real(self):
        """
        Take real part of field components.
        """
        for field in "Ex Ey Ez Bx By Bz".split():
            setattr(self, field, np.real(getattr(self, field)))


    @abstractmethod
    def calculate_field(self, t, E_out=None, B_out=None, **kwargs):
        """
        Calculates fields for a given time step.

        Parameters
        ----------
        t : float
            The time step at which to calculate the fields.
        E_out : array-like, optional
            Output array for the electric field.
        B_out : array-like, optional
            Output array for the magnetic field.
        **kwargs : dict
            Additional parameters for field calculation.

        Returns
        -------
        E_out, B_out : array-like
            The electric and magnetic fields at the given time step.
        """
        ...


class ExplicitField(Field):
    """
    For fields with an analytic formula known in the whole space at 
    least in the focus (t=0).

    One time step (e.g., focus at t=0) can serve as a model field
    for the calculation of Maxwell coefficients and Maxwell propagation
    to other time steps.

    Parameters
    ----------
    grid : quvac.grid.GridXYZ
        Spatial and spectral grid.

    Attributes
    ----------
    grid_xyz : quvac.grid.GridXYZ
        The spatial and spectral grid.
    Ef : list of complex numpy.ndarray
        The electric field in Fourier space.
    Ef_fftw : list of pyfftw.FFTW
        The FFTW objects for the electric field.
    a1 : complex numpy.ndarray
        The a1 spectral coefficient.
    a2 : complex numpy.ndarray
        The a2 spectral coefficient.
    """

    def __init__(self, field_params, grid):
        super().__init__(field_params, grid)

    def _allocate_fft(self):
        """
        Allocates memory for FFT calculations.
        """
        self.Ef = [
            pyfftw.zeros_aligned(self.grid_shape, dtype="complex128") for _ in range(3)
        ]
        # pyfftw scheme
        self.Ef_fftw = [
            pyfftw.FFTW(
                a,
                a,
                axes=(0, 1, 2),
                direction="FFTW_FORWARD",
                flags=("FFTW_MEASURE",),
                threads=1,
            )
            for a in self.Ef
        ]

    def _check_energy_kspace(self):
        # Fix energy
        W_upd = get_field_energy_kspace(
            self.a1, self.a2, self.kabs, self.dVk, mode="without 1/k"
        )
        _logger.info(f"    Energy after projection in k-space: {W_upd:.3f} J")

        if W_upd > 0:
            self.a1 *= np.sqrt(self.W / W_upd)
            self.a2 *= np.sqrt(self.W / W_upd)
        else:
            self.a1 *= 0
            self.a2 *= 0

        W_corrected = get_field_energy_kspace(
            self.a1, self.a2, self.kabs, self.dVk, mode="without 1/k"
        )
        _logger.info(f'    Energy after "correction":          {W_corrected:.3f} J')
        

    def get_a12(self, t0=None):
        """
        Calculates the a1 and a2 coefficients at a given time step.

        Parameters
        ----------
        t0 : float, optional
            The time step at which to calculate the coefficients. If None, uses self.t0.

        Returns
        -------
        a1 : numpy.ndarray
            The a1 spectral coefficient.
        a2 : numpy.ndarray
            The a2 spectral coefficient.

        Notes
        -----
        After the projection of the field on spectral coefficients, its energy
        is corrected to the desired value.
        """
        t0 = t0 if t0 is not None else self.t0
        self._allocate_fft()
        self.calculate_field(t0, E_out=self.Ef, mode="complex")

        for idx in range(3):
            self.Ef_fftw[idx].execute()

        # Calculate a1, a2 coefficients
        Efx, Efy, Efz = self.Ef

        self.a1 = ne.evaluate(
            "dV * (e1x*Efx + e1y*Efy + e1z*Efz)", global_dict=self.__dict__
        )
        self.a2 = ne.evaluate(
            "dV * (e2x*Efx + e2y*Efy + e2z*Efz)", global_dict=self.__dict__
        )

        self._check_energy_kspace()

        del self.Ef, self.Ef_fftw
        return self.a1, self.a2


class SpectralField(Field):
    """
    For fields with analytic formula to its spectrum. From this
    formula the spectral coefficients are explicitly defined and
    the field can be passed to Maxwell solver.

    Parameters
    ----------
    grid : quvac.grid.GridXYZ
        Spatial and spectral grid.

    Attributes
    ----------
    grid_xyz : quvac.grid.GridXYZ
        The spatial and spectral grid.
    a1 : complex numpy.ndarray
        The a1 spectral coefficient.
    a2 : complex numpy.ndarray
        The a2 spectral coefficient.
    """

    def __init__(self, field_params, grid):
        super().__init__(field_params, grid)

    def rotate_kgrid(self):
        """
        Rotates the k-space grid.
        """
        self.get_rotation()
        axes = "kx_rotated ky_rotated kz_rotated".split()
        kx_, ky_, kz_ = np.meshgrid(*self.kgrid_shifted, indexing='ij')
        for i, ax in enumerate(axes):
            mx, my, mz = self.rotation_bwd_m[i, :]
            new_ax = ne.evaluate(
                "mx*kx_ + my*ky_ + mz*kz_", global_dict=self.__dict__
            )
            setattr(self, ax, new_ax)
            
        self.kabs_rotated = ne.evaluate(
            "sqrt(kx_rotated**2 + ky_rotated**2 + kz_rotated**2)",
            global_dict=self.__dict__
        )
    
    def rotate_vector_potential_back(self):
        """
        Rotate the fields back to the original coordinate frame.

        Returns
        -------
        tuple of array-like
            The rotated vector potential.
        """
        A = [np.zeros(self.grid_shape, dtype=np.complex128) for _ in range(3)]

        out_dtype = config.CDTYPE
        # Transform to the original coordinate frame
        for i,Ai in enumerate(A):
            mx, my, mz = self.rotation_m[i, :]
            Ai += ne.evaluate("mx*vector_potential",
                              global_dict=self.__dict__).astype(out_dtype)
        return A

    def _check_energy_kspace(self):
        # Fix energy
        W_upd = get_field_energy_kspace(
            self.a1, self.a2, self.kabs, self.dVk, mode="without 1/k"
        )
        _logger.info(f"    Energy after projection in k-space: {W_upd:.3f} J")

        if W_upd > 0:
            self.a1 *= np.sqrt(self.W / W_upd)
            self.a2 *= np.sqrt(self.W / W_upd)
        else:
            self.a1 *= 0
            self.a2 *= 0

        W_corrected = get_field_energy_kspace(
            self.a1, self.a2, self.kabs, self.dVk, mode="without 1/k"
        )
        _logger.info(f'    Energy after "correction":          {W_corrected:.3f} J')

    def get_a12(self, t0=None):
        """
        Calculates the a1 and a2 coefficients at a given time step.

        Parameters
        ----------
        t0 : float, optional
            The time step at which to calculate the coefficients. If None, uses self.t0.

        Returns
        -------
        a1 : numpy.ndarray
            The a1 spectral coefficient.
        a2 : numpy.ndarray
            The a2 spectral coefficient.

        Notes
        -----
        After the projection of the field on spectral coefficients, its energy
        is corrected to the desired value.
        """
        t0 = t0 if t0 is not None else self.t0

        self.a1 = ne.evaluate(
            "(e1x*Ax + e1y*Ay + e1z*Az)", global_dict=self.__dict__
        )
        self.a2 = ne.evaluate(
            "(e2x*Ax + e2y*Ay + e2z*Az)", global_dict=self.__dict__
        )

        self._check_energy_kspace()

        return self.a1, self.a2


class FieldFromFile(Field):
    """
    For fields that are loaded from file.

    Not implemented yet.
    """



"""
Collection of tests for utils.
"""
import os

import numpy as np
import pytest

from quvac.utils import (
    find_classes_in_package,
    format_memory,
    format_time,
    load_wisdom,
    read_yaml,
    round_to_n,
    save_wisdom,
    write_yaml,
)

TEST_YAML = {
    "param_1": 1,
    "param_10": 10,
}


@pytest.fixture(scope="session")
def get_tmp_path(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("data")
    return tmp_path


def test_yaml(tmp_path):
    yaml_path = tmp_path / "test.yml"
    write_yaml(yaml_path, TEST_YAML)
    yaml_from_file = read_yaml(yaml_path)
    assert yaml_from_file == TEST_YAML


def test_wisdom(tmp_path):
    ini_path = tmp_path / "ini.yml"
    save_wisdom(ini_path)

    wisdom_path = os.path.dirname(ini_path)
    wisdom_file = os.path.join(wisdom_path, "fftw-wisdom")
    wisdom = load_wisdom(wisdom_file)
    print(wisdom)
    assert len(wisdom) > 0, "Wisdom file should be not empty."


@pytest.mark.parametrize("t, expected", [
    (48, "   48.00 s"),
    (61, "  1 min 1.00 s"),
    (3661, " 1 h 1 min 1.00 s"),
    ])
def test_format_time(t, expected):
    t_str = format_time(t)
    assert t_str == expected


@pytest.mark.parametrize("mem, expected", [
    (128, "128.00 KB"),
    (1025, "1.00 MB"),
    (1048577, "1.00 GB"),
])
def test_format_memory(mem, expected):
    mem_str = format_memory(mem)
    assert mem_str == expected


def test_find_classes():
    cls_names = find_classes_in_package("quvac")
    assert len(cls_names) > 0, "Quvac package has classes."


@pytest.mark.parametrize("num, number_of_digits, expected", [
    (12345, 1, 12000),
    (12345, 3, 12340),
    (0.000345, 1, 0.00034),
])
def test_round(num, number_of_digits, expected):
    num_rounded = round_to_n(num, number_of_digits)
    assert np.isclose(num_rounded, expected)




"""
Useful plotting functions.
"""

import os
from pathlib import Path

import numpy as np
from scipy.constants import pi

try:
    import matplotlib.colors as mcolors
    import matplotlib.pyplot as plt
    MATPLOTLIB_AVAILABLE = True
except ModuleNotFoundError:
    MATPLOTLIB_AVAILABLE = False


def _check_matplotlib():
    if not MATPLOTLIB_AVAILABLE:
        raise ImportError("Matplotlib is required for this function.")

def save_fig(save_path, fig_name):
    """
    Save figure to file as png and pdf.

    Parameters
    ----------
    save_path : str
        Path to the directory where the figure will be saved.
    fig_name : str
        Name of the figure file.
    """
    _check_matplotlib()
    if save_path is not None:
        Path(save_path).mkdir(parents=True, exist_ok=True)
        name = os.path.join(save_path, fig_name)
        plt.savefig(f"{name}.png", bbox_inches='tight')
        plt.savefig(f"{name}.pdf", bbox_inches='tight')


def pi_formatter(x, pos):
    """
    Format axis ticks as multiples of pi.

    Parameters
    ----------
    x : float
        Tick value.
    pos : int
        Tick position.

    Returns
    -------
    str
        Formatted tick label.
    """
    fractions = {0: "0", pi/8: r"$\frac{\pi}{8}$", pi/4: r"$\frac{\pi}{4}$", 
                 3*pi/8: r"$\frac{3\pi}{8}$", pi/2: r"$\frac{\pi}{2}$",
                 5*pi/8: r"$\frac{5\pi}{8}$", 3*pi/4: r"$\frac{3\pi}{4}$",
                 7*pi/8: r"$\frac{7\pi}{8}$", pi: r"$\pi$",
                 9*pi/8: r"$\frac{9\pi}{8}$", 5*pi/4: r"$\frac{5*\pi}{4}$",
                 11*pi/8: r"$\frac{11\pi}{8}$", 3*pi/2: r"$\frac{3\pi}{2}$",
                 13*pi/8: r"$\frac{13\pi}{8}$", 7*pi/4: r"$\frac{7\pi}{4}$",
                 15*pi/8: r"$\frac{15\pi}{8}$", 2*pi: r"$2\pi$"}
    return fractions.get(x, f"${x/np.pi:.2g}\\pi$")


def plot_roi(ax, x0, y0, dx, dy, line_kwargs):
    """
    Plot a rectangular region of interest (ROI) on a given axis.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axis on which to plot the ROI.
    x0 : float
        The x-coordinate of the center of the ROI.
    y0 : float
        The y-coordinate of the center of the ROI.
    dx : float
        Half the width of the ROI.
    dy : float
        Half the height of the ROI.
    line_kwargs : dict
        Keyword arguments to customize the appearance of the ROI lines 
        (e.g., color, linestyle).

    Returns
    -------
    matplotlib.axes.Axes
        The axis with the plotted ROI.

    Notes
    -----
    The ROI is represented as a rectangle centered at (x0, y0) with width 2*dx and 
    height 2*dy.
    """
    x_left, x_right = x0-dx, x0+dx
    y_top, y_bottom = y0-dy, y0+dy
    pts = [(x_right,y_top),(x_left,y_top),(x_left,y_bottom),
           (x_right,y_bottom),(x_right,y_top)]
    for pt1,pt2 in zip(pts[:-1],pts[1:],strict=True):
        ax.plot([pt1[0],pt2[0]], [pt1[1],pt2[1]], **line_kwargs)
    return ax


def plot_mollweide(fig, ax, phi, theta, data, cmap='coolwarm', norm=None):
    """
    Plot data on a Mollweide projection.

    Parameters
    ----------
    fig : matplotlib.figure.Figure
        Figure object.
    ax : matplotlib.axes.Axes
        Axes object.
    phi : numpy.ndarray
        Array of azimuthal angles.
    theta : numpy.ndarray
        Array of polar angles.
    data : numpy.ndarray
        Data to be plotted.
    cmap : str, optional
        Colormap, by default 'coolwarm'.
    norm : matplotlib.colors.Normalize, optional
        Normalization for the colormap, by default None.

    Returns
    -------
    ax : matplotlib.axes.Axes
        Axes object with the plot.
    """
    _check_matplotlib()
    theta_ = theta - np.pi/2
    # flip theta axis so mollweide plot shows usual sphere surface
    theta_ = theta_[::-1]
    phi_ = phi - np.pi
    phi_mesh, theta_mesh = np.meshgrid(phi_, theta_)
    
    im = ax.pcolormesh(phi_mesh, theta_mesh, data, cmap=cmap,
                       shading='gouraud', rasterized=True, norm=norm)
    cbar = fig.colorbar(im, ax=ax, shrink=0.5)

    ax.set_xticks([-2, -1, 0, 1, 2])
    ax.set_yticks([-1.5, -1, -0.5, 0, 0.5, 1, 1.5])
    xtick_labels = np.linspace(60, 360, 5, endpoint=False, dtype=int)
    ax.xaxis.set_ticklabels(f'${num}^{{\\circ}}$' for num in xtick_labels)
    ytick_labels = np.linspace(0, 180, 7, endpoint=True, dtype=int)[::-1]
    ax.yaxis.set_ticklabels(f'${num}^{{\\circ}}$' for num in ytick_labels)
    for item in ax.xaxis.get_ticklabels() + ax.yaxis.get_ticklabels():
        item.set_fontsize(18)
    ax.grid()
    return ax, cbar


def plot_fields(field, t, plot_keys=None, cmap='coolwarm',
                cnorm='log', norm_lim=1e-20, save_path=None):
    """
    Plot specified field components.

    Plotted figures:
    1. (x,y) and (x,z) profiles at the focus.
    2. x, y, z slices through the focus.

    Parameters
    ----------
    field : quvac.field.Field
        Field object containing the field data.
    t : float
        Time at which to plot the field.
    plot_keys : list of str, optional
        List of field components to plot, by default None.
    cmap : str, optional
        Colormap, by default 'coolwarm'.
    cnorm : str, optional
        Normalization type, by default 'log'.
    norm_lim : float, optional
        Normalization limit, by default 1e-20.
    save_path : str, optional
        Path to save the plots, by default None.

    Returns
    -------
    field_comps : dict
        Dictionary containing the field components.
    """
    _check_matplotlib()
    Nxyz = field.grid_xyz.grid_shape
    E_out, B_out = [[np.zeros(Nxyz, dtype=np.complex128) for _ in range(3)]
                    for _ in range(2)]
    E, B = field.calculate_field(t=t, E_out=E_out, B_out=B_out)
    E, B = [np.real(Ex) for Ex in E], [np.real(Bx) for Bx in B]
    nx, ny, nz = Nxyz
    x, y, z = [ax*1e6 for ax in field.grid_xyz.grid]

    intensity = (E[0]**2 + E[1]**2 + E[2]**2 + B[0]**2 + B[1]**2 + B[2]**2)/2
    field_comps = {
        "Ex": E[0],
        "Ey": E[1],
        "Ez": E[2],
        "Bx": B[0],
        "By": B[1],
        "Bz": B[2],
        "Intensity": intensity,
    }

    plot_keys = plot_keys if plot_keys is not None else field_comps.keys()
    # 1st plot: xy and xz profiles at focus for given components
    n_rows = len(plot_keys)
    n_cols = 2

    plt.figure(figsize=(12, 5*n_rows), layout="constrained")
    for i,key in enumerate(plot_keys):
        if key == "Intensity":
            cmap = "inferno"

        ax_bottom, ax_top = [y, z], [x, x]
        x_labels = "y z".split()
        comps = [field_comps[key][:, :, nz//2], field_comps[key][:, ny//2, :]]

        for j,comp in enumerate(comps):
            norm = None
            if key == "Intensity" and cnorm == "log":
                norm = mcolors.LogNorm(vmin=comp.max()*norm_lim, vmax=comp.max())
            ax = plt.subplot(n_rows, n_cols, i*n_cols+j+1)
            im = plt.pcolormesh(ax_bottom[j], ax_top[j], comp, shading=None,
                                rasterized=True, cmap=cmap, norm=norm)
            plt.xlabel(f"{x_labels[j]} [$\\mu$m]")
            if j == 0:
                plt.ylabel("x [$\\mu$m]")
            plt.title(f"{key} at t={t*1e15:.1f} fs")

            plt.colorbar(im, shrink=0.6)

            ax.set_aspect('equal')
    save_fig(save_path, "field_profiles_focus")
    plt.show()

    # 2nd plot: x, y, z slices through focus for given components
    n_rows = len(plot_keys)
    n_cols = 3
    axs = [x, y, z]
    axs_names = ["x", "y", "z"]

    plt.figure(figsize=(18, 5*n_rows), layout="constrained")
    for i,key in enumerate(plot_keys):
        comp = field_comps[key]
        slices = [comp[:, ny//2, nz//2],
                  comp[nx//2, :, nz//2],
                  comp[nx//2, ny//2, :]]
        for j,slc in enumerate(slices):
            plt.subplot(n_rows, n_cols, i*n_cols+j+1)
            plt.plot(axs[j], np.abs(slc))
            I0 = np.abs(slc).max()
            if norm_lim:
                plt.ylim(I0*norm_lim, 3*I0)
            plt.yscale("log")
            plt.xlabel(f"{axs_names[j]} [$\\mu$m]")
            if j == 0:
                plt.ylabel(key)
            plt.title(f"{axs_names[j]} slice")
    save_fig(save_path, "field_slices_focus")
    plt.show()

    return field_comps




"""
Calculation of vacuum emission integral (box diagram, F^4).

Currently supports:

1. Calculation of total vacuum emission signal for given field configuration. 
All fields are treated as external.

2. Separation of fields into pump and probe with subsequent calculation of
probe channel signal.

.. note::
    For details on implementation check out :ref:`implementation` section.
"""

import os
from pathlib import Path
import time

import numexpr as ne
import numpy as np
import pyfftw
from scipy.constants import alpha, c, e, hbar, m_e, pi

from quvac import config

BS = m_e**2 * c**2 / (hbar * e)  # Schwinger magnetic field


class VacuumEmission:
    """
    Calculator of Vacuum Emission amplitude from given fields

    Parameters
    ----------
    field : quvac.Field
        External fields.
    grid : quvac.grid.GridXYZ
        Spatial and spectral grid.
    nthreads : int, optional
        Number of threads to use for calculations. If not provided, 
        defaults to the number of CPU cores.
    channels : bool, optional
        Whether to calculate a particular channel in vacuum emission amplitude. 
        Default is False.

    """

    def __init__(self, field, grid, nthreads=None, channels=False):
        self.field = field
        self.grid_xyz = grid
        # Update local dict with variables from GridXYZ class
        self.__dict__.update(self.grid_xyz.__dict__)
        self.channels = channels

        self.c = c
        self.nthreads = nthreads if nthreads else os.cpu_count()

        # Define symbolic expressions to evaluate later
        self.F_expr = "(Bx**2 + By**2 + Bz**2 - Ex**2 - Ey**2 - Ez**2)/2"
        self.G_expr = "-(Ex*Bx + Ey*By + Ez*Bz)"

        self.F, self.G = [
            np.zeros(self.grid_shape, dtype=config.FDTYPE) for _ in range(2)
        ]

        if not self.channels:
            self.U1 = [f"(4*E{ax}*F + 7*B{ax}*G)" for ax in "xyz"]
            self.U2 = [f"(4*B{ax}*F - 7*E{ax}*G)" for ax in "xyz"]
        else:
            self._define_channel_variables()

        self.I_ij = {
            f"{i}{j}": f"(e{i}x*U{j}_acc_x + e{i}y*U{j}_acc_y + e{i}z*U{j}_acc_z)"
            for i in range(1, 3)
            for j in range(1, 3)
        }
        for key, val in self.I_ij.items():
            self.__dict__[f"I_{key}_expr"] = val

    def _define_channel_variables(self):
        """
        Define variables for channel-separated signal (linear in probe).
        """
        self.F_B_Bp_expr = "(Bx*Bpx + By*Bpy + Bz*Bpz - Ex*Epx - Ey*Epy - Ez*Epz)"
        self.G_Ep_B_expr = "-(Epx*Bx + Epy*By + Epz*Bz)"
        self.G_E_Bp_expr = "-(Ex*Bpx + Ey*Bpy + Ez*Bpz)"

        self.F_B_Bp, self.G_Ep_B, self.G_E_Bp = [
            np.zeros(self.grid_shape, dtype=config.FDTYPE) for _ in range(3)
        ]

        self.U1 = [
            f"(4*(Ep{ax}*F + E{ax}*F_B_Bp) + 7*(Bp{ax}*G + B{ax}*(G_Ep_B + G_E_Bp)))"
            for ax in "xyz"
        ]
        self.U2 = [
            f"(4*(Bp{ax}*F + B{ax}*F_B_Bp) - 7*(Ep{ax}*G + E{ax}*(G_Ep_B + G_E_Bp)))"
            for ax in "xyz"
        ]

    def _allocate_fields(self):
        """
        Allocate memory for field calculations.
        """
        self.E_out = [np.zeros(self.grid_shape, dtype=config.CDTYPE) for _ in range(3)]
        self.B_out = [np.zeros(self.grid_shape, dtype=config.CDTYPE) for _ in range(3)]
        if self.channels:
            self.E_probe = [
                np.zeros(self.grid_shape, dtype=config.CDTYPE) for _ in range(3)
            ]
            self.B_probe = [
                np.zeros(self.grid_shape, dtype=config.CDTYPE) for _ in range(3)
            ]

    def _allocate_result_arrays(self):
        """
        Allocate memory for result arrays.
        """
        self.U1_acc = [np.zeros(self.grid_shape, dtype=config.CDTYPE) for _ in range(3)]
        self.U2_acc = [np.zeros(self.grid_shape, dtype=config.CDTYPE) for _ in range(3)]
        self.U1_acc_x, self.U1_acc_y, self.U1_acc_z = self.U1_acc
        self.U2_acc_x, self.U2_acc_y, self.U2_acc_z = self.U2_acc

    def _allocate_fft(self):
        """
        Allocate memory for FFT calculations.
        """
        self.tmp = [
            pyfftw.zeros_aligned(self.grid_shape, dtype=config.CDTYPE) for _ in range(3)
        ]
        # Add number of threads
        self.tmp_fftw = [
            pyfftw.FFTW(
                a,
                a,
                axes=(0, 1, 2),
                direction="FFTW_FORWARD",
                flags=("FFTW_MEASURE",),
                threads=self.nthreads,
            )
            for a in self.tmp
        ]
        self.prefactor = np.zeros(self.grid_shape, dtype="complex128")

        self.prefactor_dict = {
            "kabs": self.kabs,
            "c": c,
        }

        self.U_dict = {"F": self.F, "G": self.G}

    def _free_resources(self):
        """
        Free allocated resources.
        """
        del self.E_out, self.B_out
        del self.tmp, self.tmp_fftw

    def calculate_one_time_step(self, t, weight=1):
        """
        Calculate the field and U terms (integrand) for one time step.
        """
        # Calculate fields
        self._allocate_fields()

        if not self.channels:
            self.field.calculate_field(t, E_out=self.E_out, B_out=self.B_out)
        else:
            self.field.calculate_field(
                t,
                E_probe=self.E_probe,
                B_probe=self.B_probe,
                E_pump=self.E_out,
                B_pump=self.B_out,
            )
            Epx, Epy, Epz = [E.real for E in self.E_probe]
            Bpx, Bpy, Bpz = [B.real for B in self.B_probe]

        Ex, Ey, Ez = [E.real for E in self.E_out]
        Bx, By, Bz = [B.real for B in self.B_out]
        self.U_dict.update({"Ex": Ex, "Ey": Ey, "Ez": Ez, "Bx": Bx, "By": By, "Bz": Bz})

        ne.evaluate(self.F_expr, out=self.F)
        ne.evaluate(self.G_expr, out=self.G)

        if self.channels:
            ne.evaluate(self.F_B_Bp_expr, out=self.F_B_Bp)
            ne.evaluate(self.G_Ep_B_expr, out=self.G_Ep_B)
            ne.evaluate(self.G_E_Bp_expr, out=self.G_E_Bp)
            self.U_dict.update({"Epx": Epx, "Epy": Epy, "Epz": Epz,
                                "Bpx": Bpx, "Bpy": Bpy, "Bpz": Bpz,
                                "F_B_Bp": self.F_B_Bp,
                                "G_Ep_B": self.G_Ep_B,
                                "G_E_Bp": self.G_E_Bp,})

        # Evaluate U1 and U2 expressions
        self.prefactor_dict.update({"t": t})
        ne.evaluate(
            "exp(1j*kabs*c*t)", local_dict=self.prefactor_dict, out=self.prefactor
        )
        for U_key, U_expr in zip(["U1_acc", "U2_acc"], [self.U1, self.U2]): # noqa: B905
            for i, expr in enumerate(U_expr):
                U_acc = getattr(self, U_key)[i]
                ne.evaluate(expr, global_dict=self.U_dict, out=self.tmp[i])
                self.tmp_fftw[i].execute()

                U_res = ne.evaluate(
                    "U_acc + U*prefactor*dt*dV",
                    global_dict={
                        "U_acc": U_acc,
                        "U": self.tmp[i],
                        "prefactor": self.prefactor,
                        "dt": self.dt,
                        "dV": self.dV,
                    },
                )
                U_acc[:] = U_res.astype(config.CDTYPE)

    def calculate_time_integral(self, t_grid, integration_method="trapezoid"):
        """
        Calculate the time integral.
        """
        self.dt = t_grid[1] - t_grid[0]
        if integration_method == "trapezoid":
            # end_pts = (0, len(t_grid) - 1)
            for _, t in enumerate(t_grid):
                # weight = 0.5 if i in end_pts else 1.
                weight = 1
                self.calculate_one_time_step(t, weight=weight)
        else:
            err_msg = (
                "integration_method should be one of ['trapezoid'] but you "
                f"passed {integration_method}"
            )
            raise NotImplementedError(err_msg)

    def calculate_amplitudes(
        self, t_grid, integration_method="trapezoid", save_path=None
    ):
        """
        Calculate the vacuum emission amplitudes and save the result.
        """
        # Allocate resources
        self._allocate_result_arrays()
        self._allocate_fft()

        time_integral_start = time.perf_counter()
        self.calculate_time_integral(t_grid, integration_method)
        time_integral_end = time.perf_counter()
        time_integral = time_integral_end - time_integral_start
        self._free_resources()

        # Results should be in U1_acc and U2_acc
        dims = 1 / BS**3 * m_e**2 * c**3 / hbar**2
        prefactor = -1j * np.sqrt(alpha * self.kabs) / (2 * pi) ** 1.5 / 45 * dims # noqa: F841
        # Next time need to be careful with f-strings and brackets
        self.S1 = ne.evaluate(
            f"prefactor * ({self.I_11_expr} - {self.I_22_expr})",
            global_dict=self.__dict__,
        ).astype(config.CDTYPE)
        self.S2 = ne.evaluate(
            f"prefactor * ({self.I_12_expr} + {self.I_21_expr})",
            global_dict=self.__dict__,
        ).astype(config.CDTYPE)
        # Save amplitudes
        if save_path:
            self.save_amplitudes(save_path)
        return time_integral

    def save_amplitudes(self, save_path):
        """
        Save the calculated amplitudes to a file.
        """
        Path(os.path.dirname(save_path)).mkdir(parents=True, exist_ok=True)
        data = {
            "x": self.grid[0],
            "y": self.grid[1],
            "z": self.grid[2],
            "S1": self.S1,
            "S2": self.S2,
        }
        np.savez(save_path, **data)



"""
Usefult functions and default strings for logs.
"""
from quvac.utils import format_memory, format_time

# Warning when total signal on spherical grid differs from
# cartesian grid one
sph_interp_warn = """WARNING:
{:s} signal on cartesian and spherical grid differ by more than 1%:
    N total (xyz): {:.3f}
    N total (sph): {:.3f}
"""

# timings structure
performance_str = """
Timings:
====================================================
Field setup:               {:>15s}
Vacem setup:               {:>15s}
Amplitudes calculation:    {:>15s}
    Per time step:         {:>15s}
Postprocess:               {:>15s}
----------------------------------------------------
Total:                     {:>15s}
====================================================

Memory (max usage):
====================================================
Amplitudes calculation:    {:>15s}
Total:                     {:>15s}
====================================================
"""

# perf string for postprocess
performance_post_str = """
Postprocess:
====================================================
Time:                      {:>15s}
----------------------------------------------------
Memory (max usage):        {:>15s}
====================================================
"""

# timings for parallel
performance_parallel_str = """
Timings:
====================================================
Run jobs:                  {:>15s}
Postprocess:               {:>15s}
----------------------------------------------------
Total:                     {:>15s}
====================================================

Memory (max usage):
====================================================
Running jobs:              {:>15s}
Total:                     {:>15s}
====================================================
"""

# grid params
grid_str = """
Grid:
====================================================
Space

Number of points (x,y,z):  {:>25}
Box for x axis:            {:>25}
Box for y axis:            {:>25}
Box for z axis:            {:>25}
----------------------------------------------------
Time

Number of points:          {:>25}
Box:                       {:>25}
====================================================
"""


postprocess_str = """
Postprocess parameters
====================================================
Perp polarization type:           {:>15}
Perp field idx:                   {:>15}
Signal on spherical grid:         {:>15}
Discernible signal:               {:>15}
====================================================
"""

test_run_str = """
====================================================
Time per iteration:               {:>15s}
Overhead time:                    {:>15s}  
Estimated time (full run):        {:>15s}
====================================================
"""


simulation_start_str = """
####################################################
Start of simulation:    {:>20}
####################################################
"""


simulation_end_str = """
####################################################
End of simulation:      {:>20}
####################################################
"""


def get_grid_params(grid_xyz, grid_t):
    """
    Get formatted grid parameters for logging.

    Parameters
    ----------
    grid_xyz : quvac.grid.GridXYZ
        The spatial grid object.
    grid_t : numpy.ndarray
        The temporal grid array.

    Returns
    -------
    grid_print : str
        Formatted string with grid parameters.
    """
    nx, ny, nz = grid_xyz.grid_shape
    grid_xyz_size = f"({nx}, {ny}, {nz})"
    x, y, z = grid_xyz.grid
    x_start, x_end = x[0] * 1e6, x[-1] * 1e6
    x_box = f"({x_start:.2f}, {x_end:.2f}) micron"
    y_start, y_end = y[0] * 1e6, y[-1] * 1e6
    y_box = f"({y_start:.2f}, {y_end:.2f}) micron"
    z_start, z_end = z[0] * 1e6, z[-1] * 1e6
    z_box = f"({z_start:.2f}, {z_end:.2f}) micron"

    grid_t_size = len(grid_t)
    t_start, t_end = grid_t[0] * 1e15, grid_t[-1] * 1e15
    t_box = f"({t_start:.2f}, {t_end:.2f}) fs"

    grid_print = grid_str.format(grid_xyz_size, x_box, y_box, z_box, grid_t_size, t_box)
    return grid_print


def get_postprocess_info(postprocess_params):
    """
    Get formatted postprocess parameters for logging.

    Parameters
    ----------
    postprocess_params : dict
        Dictionary containing postprocess parameters.

    Returns
    -------
    postprocess_print : str
        Formatted string with postprocess parameters.
    """
    calculate_spherical = postprocess_params.get("calculate_spherical", False)
    # spherical_params = postprocess_params.get("spherical_params", {})
    calculate_discernible = postprocess_params.get("calculate_discernible", False)
    perp_type = postprocess_params.get("perp_polarization_type", "optical axis")
    perp_field_idx = postprocess_params.get("perp_field_idx", 1)

    postprocess_print = postprocess_str.format(
        str(perp_type), str(perp_field_idx), str(calculate_spherical),
        str(calculate_discernible)
    )
    return postprocess_print


def get_performance_stats(perf_stats):
    """
    Get formatted performance statistics for logging.

    Parameters
    ----------
    perf_stats : dict
        Dictionary containing performance statistics.

    Returns
    -------
    perf_print : str
        Formatted string with performance statistics.
    """
    timings = perf_stats["timings"]
    timings = {
        "field_setup": timings["field_setup"] - timings["start"],
        "vacem_setup": timings["vacem_setup"] - timings["field_setup"],
        "amplitudes": timings["amplitudes"] - timings["vacem_setup"],
        "postprocess": timings["postprocess"] - timings["amplitudes"],
        "per_iteration": timings["per_iteration"],
        "total": timings["postprocess"] - timings["start"],
    }
    timings = {k: format_time(t) for k, t in timings.items()}
    memory = {k: format_memory(m) for k, m in perf_stats["memory"].items()}
    perf_print = performance_str.format(
        timings["field_setup"],
        timings["vacem_setup"],
        timings["amplitudes"],
        timings["per_iteration"],
        timings["postprocess"],
        timings["total"],
        memory["maxrss_amplitudes"],
        memory["maxrss_total"],
    )
    return perf_print


def get_postprocess_stats(perf_stats):
    """
    Get formatted postprocess performance statistics for logging.

    Parameters
    ----------
    perf_stats : dict
        Dictionary containing postprocess performance statistics.

    Returns
    -------
    perf_print : str
        Formatted string with postprocess performance statistics.
    """
    timings = perf_stats['timings']
    timings = {
        'postprocess': timings["postprocess"] - timings["start"]
    }
    timings = {k: format_time(t) for k, t in timings.items()}
    memory = {k: format_memory(m) for k, m in perf_stats["memory"].items()}
    perf_print = performance_post_str.format(
        timings['postprocess'],
        memory["maxrss_total"]
    )
    return perf_print


def get_parallel_performance_stats(perf_stats):
    """
    Get formatted parallel performance statistics for logging.

    Parameters
    ----------
    perf_stats : dict
        Dictionary containing parallel performance statistics.

    Returns
    -------
    perf_print : str
        Formatted string with parallel performance statistics.
    """
    timings = perf_stats["timings"]
    timings = {
        "jobs": timings["jobs"] - timings["start"],
        "postprocess": timings["postprocess"] - timings["jobs"],
        "total": timings["postprocess"] - timings["start"],
    }
    timings = {k: format_time(t) for k, t in timings.items()}
    memory = {k: format_memory(m) for k, m in perf_stats["memory"].items()}
    perf_print = performance_parallel_str.format(
        timings["jobs"],
        timings["postprocess"],
        timings["total"],
        memory["maxrss_jobs"],
        memory["maxrss_total"],
    )
    return perf_print


def get_test_timings(timings, Nt, expected_timesteps):
    """
    Get formatted test run timings for logging.

    Parameters
    ----------
    timings : dict
        Dictionary containing timing information.
    Nt : int
        Number of iterations.
    expected_timesteps : int
        Expected number of timesteps.

    Returns
    -------
    test_run_str_print : str
        Formatted string with test run timings.
    """
    time_overhead = timings['amplitudes'] - timings['field_setup'] - timings['integral']
    time_per_iteration = timings['integral'] / Nt
    expected_time = time_per_iteration * expected_timesteps
    time_total = time_overhead + expected_time
    test_run_str_print = test_run_str.format(
        format_time(time_per_iteration),
        format_time(time_overhead),
        format_time(time_total),
    )
    return test_run_str_print


#!/usr/bin/env python3
"""
Script to launch Vacuum Emission simulation IN PARALLEL 
(using submitit for semi-slurm submission type), do postprocessing
and measure performance.

Usage:

.. code-block:: bash

    python simulation_parallel.py -i <input>.yaml -o <output_dir> 
    --wisdom <wisdom_file>
"""
from copy import deepcopy
import logging
import os
from pathlib import Path
import time

import numpy as np
import submitit

from quvac.config import DEFAULT_SUBMITIT_PARAMS
from quvac.grid import setup_grids
from quvac.log import (
    get_parallel_performance_stats,
    simulation_end_str,
    simulation_start_str,
)
from quvac.simulation import (
    get_dirs,
    parse_args,
    postprocess_simulation,
    quvac_simulation,
)
from quvac.utils import get_maxrss, read_yaml, write_yaml

_logger = logging.getLogger("simulation")


def create_ini_files_for_parallel(ini_config, grid_xyz, grid_t, n_jobs, save_path):
    """
    Create initialization files for parallel jobs.

    Parameters
    ----------
    ini_config : dict
        Dictionary containing the initialization configuration.
    grid_xyz : quvac.grid.GridXYZ
        The spatial grid object.
    grid_t : numpy.ndarray
        The temporal grid array.
    n_jobs : int
        Number of parallel jobs.
    save_path : str
        Path to save the initialization files.

    Returns
    -------
    ini_files : list of str
        List of paths to the initialization files for each job.
    """
    Nt_total = len(grid_t)
    Nt_per_job = Nt_total // n_jobs

    ini_job = deepcopy(ini_config)
    ini_job["mode"] = "simulation"
    ini_job["postprocess"] = {}

    box_xyz = [float(-ax[0] * 2) for ax in grid_xyz.grid]
    Nxyz = [int(N) for N in grid_xyz.grid_shape]
    grid_params = {"mode": "direct", "box_xyz": box_xyz, "Nxyz": Nxyz}
    ini_job["grid"].update(grid_params)
    # Create ini files for all jobs
    ini_files = []
    for idx in range(n_jobs):
        idx_start, idx_end = Nt_per_job * idx, Nt_per_job * (idx + 1) - 1
        Nt = Nt_per_job
        if (n_jobs - 1) == idx and n_jobs > 1:
            idx_end += 1
            Nt += 1
        box_t = [float(grid_t[idx_start]), float(grid_t[idx_end])]
        grid_params_job = {
            "box_t": box_t,
            "Nt": Nt,
        }
        ini_job["grid"].update(grid_params_job)
        ini_path_job = os.path.join(save_path, f"job_{str(idx).zfill(2)}", "ini.yml")
        Path(os.path.dirname(ini_path_job)).mkdir(parents=True, exist_ok=True)
        write_yaml(ini_path_job, ini_job)
        ini_files.append(ini_path_job)
    return ini_files


def collect_results(ini_files, amplitudes_file):
    """
    Collect results from parallel jobs and combine them.

    Parameters
    ----------
    ini_files : list of str
        List of paths to the initialization files for each job.
    amplitudes_file : str
        Path to save the combined amplitudes.
    """
    amplitude_files = [
        os.path.join(os.path.dirname(ini_file), "amplitudes.npz")
        for ini_file in ini_files
    ]
    amplitude_0 = np.load(amplitude_files[0])
    x, y, z = [amplitude_0[ax] for ax in "xyz"]
    S1, S2 = 0, 0
    for amplitude in amplitude_files:
        data = np.load(amplitude)
        S1 += data["S1"]
        S2 += data["S2"]
    amplitude_total = {"x": x, "y": y, "z": z, "S1": S1, "S2": S2}
    np.savez(amplitudes_file, **amplitude_total)


def quvac_simulation_parallel(
    ini_file, save_path=None, wisdom_file="wisdom/fftw-wisdom"
):
    """
    Launch a single quvac simulation for given <ini>.yaml file in parallel.

    Depending on available jobs, we split the total time interval
    into several sub-intervals and submit each sub-interval for
    calculation as a separate quvac simulation (without postprocessing).
    Then we gather all S1 and S2 in the main process and do main postprocessing.

    Parameters
    ----------
    ini_file : str
        Path to the initialization file.
    save_path : str, optional
        Path to save simulation results to, by default None.
    wisdom_file : str, optional
        Path to save pyfftw-wisdom, by default "wisdom/fftw-wisdom".
    """
    # Check that ini file and save_path exists
    files = get_dirs(ini_file, save_path, wisdom_file)

    # Load and parse ini yaml file
    ini_config = read_yaml(ini_file)
    # One can choose either to perform just simulation, just postprocess or both
    mode = ini_config.get('mode', 'simulation_postprocess')
    do_postprocess = 'postprocess' in mode

    # Setup logger
    logging.basicConfig(
        filename=files['logger'],
        filemode="w",
        encoding="utf-8",
        level=logging.DEBUG,
        format="%(message)s",
    )

    # Start time
    time_log_start = time.asctime(time.localtime())
    start_print = simulation_start_str.format(time_log_start)
    _logger.info(start_print)
    timings = {}
    timings['start'] = time.perf_counter()

    # Load and parse ini yaml file
    fields_params = ini_config["fields"]
    if isinstance(fields_params, dict):
        fields_params = list(fields_params.values())
    grid_params = ini_config["grid"]
    cluster_params = ini_config.get("cluster_params", {})

    # Get parallelization params
    n_jobs = cluster_params.get("n_jobs", 2)
    max_jobs = cluster_params.get("max_jobs", n_jobs)
    cluster = cluster_params.get("cluster", "local")
    sbatch_params = cluster_params.get("sbatch_params", DEFAULT_SUBMITIT_PARAMS)

    # Get grids
    grid_xyz, grid_t = setup_grids(fields_params, grid_params)

    ini_files = create_ini_files_for_parallel(
        ini_config, grid_xyz, grid_t, n_jobs, files['save_path']
    )

    # Create a cluster
    submitit_folder = os.path.join(files['save_path'], "submitit_logs")
    executor = submitit.AutoExecutor(folder=submitit_folder, cluster=cluster)
    if cluster == "slurm":
        executor.update_parameters(slurm_array_parallelism=max_jobs)
    executor.update_parameters(**sbatch_params)
    # else:
    #     timeout_min = sbatch_params.get("timeout_min", 5)
    #     executor.update_parameters(timeout_min=timeout_min)

    # Submit jobs
    _logger.info("MILESTONE: Submitting jobs...")
    jobs = executor.map_array(quvac_simulation, ini_files)
    _logger.info("MILESTONE: Jobs submitted, waiting for results...")

    # Wait till all jobs end
    _ = [job.result() for job in jobs]
    _logger.info("MILESTONE: Jobs are finished")

    # Collect all results
    collect_results(ini_files, files['amplitudes'])
    _logger.info("MILESTONE: Results from individual jobs are collected")
    timings['jobs'] = time.perf_counter()

    maxrss_jobs = get_maxrss()

    # Calculate spectra
    if do_postprocess:
        postprocess_simulation(ini_config, files, fields_params)

    timings['postprocess'] = time.perf_counter()
    timings['total'] = timings['postprocess'] - timings['start']

    maxrss_total = get_maxrss()

    memory = {"maxrss_jobs": maxrss_jobs, "maxrss_total": maxrss_total}
    perf_stats = {"timings": timings, "memory": memory}

    write_yaml(files['performance'], perf_stats)

    perf_print = get_parallel_performance_stats(perf_stats)
    print(perf_print)
    _logger.info(perf_print)

    print("Simulation finished!")

    # End time
    time_log_end = time.asctime(time.localtime())
    end_print = simulation_end_str.format(time_log_end)
    _logger.info(end_print)


def main_simulation_parallel():
    """
    Main function to run the simulation in parallel.
    """
    args = parse_args()
    quvac_simulation_parallel(args.input, args.output, args.wisdom)


if __name__ == "__main__":
    main_simulation_parallel()



"""
Model fields.

Currently implements:

1. Model electric/magnetic field with inhomogeneity in one direction
and infinite transverse extend in others. Adapted from [1]_.

----

.. [1] H. Gies, F. Karbstein, and N. Seegert. "Quantum reflection
   as a new signature of quantum vacuum nonlinearity." NJP 15.8 (2013): 
   083002.
"""

import numexpr as ne
from scipy.constants import pi

from quvac import config
from quvac.field.abc import Field


class EBInhomogeneity(Field):
    """
    Model inhomogeneities of electric/magnetic field.

    Field is of type B = B(z) * eB.

    Parameters
    ----------
    field_params : dict
        Dictionary containing the field parameters. Required keys are:
            - 'field_inhom': str
                Field inhomogeneity, either 'electric' or 'magnetic'.
            - 'theta' : float
                Polar angle of eB-vector (in degrees).
            - 'phi' : float
                Azimuthal angle of eB-vector (in degrees).
            - 'E0' : float
                Amplitude.
            - 'envelope_type' : str
                Envelope type, by default 'gauss'.

        Envelope specific arguments:
            For 'gauss':
                - 'w0' : float
                    Waist size.
            For 'gauss-modulated':
                - 'w0' : float
                    Waist size.
                - 'lam': float
                    Wavelength of modulation.
                - 'phase0': float, optional
                    Initial phase
    grid : quvac.grid.GridXYZ
        Spatial and grid.
    """
    def __init__(self, field_params, grid):
        super().__init__(field_params, grid)

        self.beta = getattr(self, "beta", 0.)
        self.field_inhom = getattr(self, "field_inhom", "magnetic")
        
        # create rotation matrices to transform the field orientation
        # without changing the grid
        self.rotate_coordinates(rotate_grid=False)

        self.get_envelope()
        self.E_expr = f"E0 * {self.envelope}"

        EB_keys = "Ex Ey Ez Bx By Bz".split()
        for key in EB_keys:
            setattr(self, key, 0.)

        if "W" in field_params:
            self.E0 = 1.
            self.check_energy()
    
    def get_envelope(self):
        """
        Define the envelope function for the field.

        The envelope function is defined based on the `envelope_type` parameter. 
        Currently, only the Gaussian envelope type is supported.

        Raises
        ------
        NotImplementedError
            If the `envelope_type` is not supported.
        """
        match self.envelope_type:
            case "gauss":
                self.envelope = "exp(-(z/w0)**2)"
            case "gauss-modulated":
                self.phase0 = getattr(self, "phase0", 0.)
                self.kz_m = 2*pi/self.lam
                self.envelope = "exp(-(z/w0)**2) * cos(kz_m*z + phase0)"
            case _:
                raise NotImplementedError(f"`{self.envelope_type}` envelope type"
                                          "is not supported")
            
    def check_energy(self):
        """
        Check and adjust the field energy.
        """
        self._check_energy()

        if self.modify_energy:
            self.E0 *= self.W_correction
    
    def calculate_field(self, t, E_out=None, B_out=None):
        E = ne.evaluate(self.E_expr, global_dict=self.__dict__).astype(config.FDTYPE)

        if self.field_inhom == "electric":
            self.Ez = E
        elif self.field_inhom == "magnetic":
            self.Bz = E
        else:
            raise NotImplementedError(f"`{self.field_inhom}` field inhomogeneity"
                                      "is not supported")
        
        E_out, B_out = self.rotate_fields_back(E_out, B_out, mode="real")
        return E_out, B_out



"""
Analytic expression for plane wave.
"""

import numexpr as ne
from scipy.constants import c, pi

from quvac.field.abc import ExplicitField


class PlaneWave(ExplicitField):
    """
    Analytic expression for paraxial Gaussian beam.

    Parameters
    ----------
    field_params : dict
        Dictionary containing the field parameters. Required keys are:
            - 'theta' : float
                Polar angle of k-vector (in degrees).
            - 'phi' : float
                Azimuthal angle of k-vector (in degrees).
            - 'beta' : float
                Polarization angle (in degrees).
            - 'lam' : float
                Wavelength of the pulse.
            - 'tau' : float
                Duration.
            - 'phase0' : float
                Phase delay at focus.
            - 'E0' : float, optional
                Amplitude (either E0 or W is required).
            - 'W' : float, optional
                Energy (either E0 or W is required).
            - 'l': int, optional
                Azimuthal index of the Laguerre-Gaussian mode.
    grid : quvac.grid.GridXYZ
        Spatial and grid.

    Notes
    -----
    It is possible to add OAM term to plane-wave: exp(-i*l*phi)
    """
    def __init__(self, field_params, grid):
        super().__init__(field_params, grid)

        if "E0" not in field_params:
            err_msg = ("Field params need to have either W (energy) or"
                       "E0 (amplitude) as key")
            assert "W" in field_params, err_msg
            self.E0 = 1.0e10

        self.x0, self.y0, self.z0 = getattr(self, "focus_x", [0,0,0])
        self.t0 = self.focus_t
        self.B0 = self.E0 / c
        self.k = 2.0 * pi / self.lam
        self.omega = c * self.k
        self.l = getattr(self, "l", 0)

        # Rotate coordinate grid
        self.rotate_coordinates()

        # Set up correct field amplitude
        if "W" in field_params:
            self.check_energy()

    def check_energy(self):
        """
        Check and adjust the field energy.
        """
        self._check_energy()

        if self.modify_energy:
            self.E0 *= self.W_correction
            self.B0 = self.E0 / c

    def calculate_field(self, t, E_out=None, B_out=None, mode="real"):
        """
        Calculates the electric and magnetic fields at a given time step.
        """
        phase_expr = "omega*(t-t0) - k*z + phase0"
        if self.l:
            phase_expr += " + l*arctan2(y,x)"
        self.phase = ne.evaluate(phase_expr, global_dict=self.__dict__)

        Et = ne.evaluate(
            "E0 * exp(-(2.*phase/(omega*tau))**2) * exp(-1.j*phase)",
            global_dict=self.__dict__,
        )

        self.Ex = self.By = Et.copy()
        self.Ey = self.Ez = self.Bx = self.Bz = 0.0

        if mode == "real":
            self.convert_fields_to_real()

        E_out, B_out = self.rotate_fields_back(E_out, B_out, mode)
        return E_out, B_out


"""
Basic linear Maxwell propagation class for a single field and 
a unified interface combining several Maxwell fields into one.
"""

import logging
import os

import numexpr as ne
import numpy as np
import pyfftw
from scipy.constants import c, pi

from quvac import config
from quvac.field import SPATIAL_MODEL_FIELDS
from quvac.field.abc import Field

_logger = logging.getLogger("simulation")


class MaxwellField(Field):
    """
    Basic linear Maxwell propagation class for a single field.

    For such fields, the initial field distribution (spectral coefficients)
    at a certain time step is given with an analytic expression or from a file.
    For later time steps, the field is propagated according to linear Maxwell
    equations.

    Parameters
    ----------
    grid : quvac.grid.GridXYZ
        Spatial and spectral grid.
    nthreads : int, optional
        Number of threads to use for calculations. If not provided, defaults to 
        the number of CPU cores.
    """

    def __init__(self, grid, nthreads=None):
        self.grid_xyz = grid
        self.__dict__.update(self.grid_xyz.__dict__)

        self.nthreads = nthreads if nthreads else os.cpu_count()

        self.c = c
        self.norm_ifft = self.dVk / (2.0 * pi) ** 3

        # 1st list for E, 2nd list for B
        self.EB_expr = [f"(e1{ax}*a1t + e2{ax}*a2t)" for ax in "xyz"] + [
            f"(e2{ax}*a1t - e1{ax}*a2t)" for ax in "xyz"
        ]

        self._allocate_tmp()

    def _allocate_ifft(self):
        """
        Allocate memory for inverse FFT calculations and define dictionaries
        for numexpr.evaluate().
        """
        self.a1t, self.a2t = [
            np.zeros(self.grid_shape, dtype="complex128") for _ in range(2)
        ]

        self.a_dict = {
            "kabs": self.kabs,
            "c": c,
            "t0": self.t0,
            "norm_ifft": self.norm_ifft,
            "a1": self.a1,
            "a2": self.a2,
        }

        self.EB_dict = {
            "e1x": self.e1x,
            "e1y": self.e1y,
            "e1z": self.e1z,
            "e2x": self.e2x,
            "e2y": self.e2y,
            "e2z": self.e2z,
            "a1t": self.a1t,
            "a2t": self.a2t,
        }

    def _allocate_tmp(self):
        """
        Allocate temporary memory for FFT calculations.
        """
        self.tmp = pyfftw.zeros_aligned(self.grid_shape, dtype="complex128")
        self.EB_fftw = pyfftw.FFTW(
                            self.tmp,
                            self.tmp,
                            axes=(0, 1, 2),
                            direction="FFTW_BACKWARD",
                            flags=("FFTW_MEASURE",),
                            threads=self.nthreads,
                       )

    def calculate_field(self, t, E_out=None, B_out=None):
        """
        Calculates the electric and magnetic fields at a given time step.
        """
        if E_out is None:
            E_out = [np.zeros(self.grid_shape, dtype=config.CDTYPE) for _ in range(3)]
            B_out = [np.zeros(self.grid_shape, dtype=config.CDTYPE) for _ in range(3)]

        # Calculate a1,a2 at time t
        self.a_dict.update({"t": t})
        ne.evaluate(
            "exp(-1j*kabs*c*(t-t0)) * a1 * norm_ifft",
            local_dict=self.a_dict,
            out=self.a1t,
        )
        ne.evaluate(
            "exp(-1j*kabs*c*(t-t0)) * a2 * norm_ifft",
            local_dict=self.a_dict,
            out=self.a2t,
        )

        # Calculate fourier of fields at time t and transform back to
        # spatial domain
        for idx in range(6):
            ne.evaluate(self.EB_expr[idx], local_dict=self.EB_dict, out=self.tmp)
            self.EB_fftw.execute()
            if idx < 3:
                E_out[idx][:] = self.tmp.astype(config.CDTYPE)
            else:
                B_out[idx-3][:] = self.tmp.astype(config.CDTYPE)
        return E_out, B_out


class MaxwellMultiple(MaxwellField):
    """
    Combine spectral coefficients from several fields and
    propagate them as one field.

    Parameters
    ----------
    fields : dict or list of dict
        Parameters of the fields.
    grid : quvac.grid.GridXYZ
        Spatial and spectral grid.
    nthreads : int, optional
        Number of threads to use for calculations. If not provided, defaults to the 
        number of CPU cores.

    Attributes
    ----------
    a1 : np.array
        First spectral coefficient.
    a2 : np.array
        Second spectral coefficient.
    fields : list of dict
        List of field parameter dictionaries.
    t0 : float
        Initial time for the field.
    """

    def __init__(self, fields, grid, nthreads=None):
        super().__init__(grid, nthreads)

        self.a1, self.a2 = [
            pyfftw.zeros_aligned(self.grid_shape, dtype=config.CDTYPE) for _ in range(2)
        ]
        self.fields = [fields] if isinstance(fields, dict) else fields
        for i, field in enumerate(self.fields):
            _logger.info(f"Setting up field {i+1}:")
            a1, a2 = self.get_a12_from_field(field)
            self.a1 += a1
            self.a2 += a2

        self._allocate_ifft()

    def get_a12_from_field(self, field_params):
        """
        Get the spectral coefficients a1 and a2 from the field parameters.

        Parameters
        ----------
        field_params : dict
            Dictionary containing the parameters for the field.

        Returns
        -------
        tuple of np.array
            The spectral coefficients a1 and a2.

        Raises
        ------
        NotImplementedError
            If the field type is not supported.
        """
        field_type = field_params["field_type"]
        if field_type in SPATIAL_MODEL_FIELDS:
            cls = SPATIAL_MODEL_FIELDS[field_type]
            _logger.info(f"    {field_type}: {cls.__name__}")
            ini_field = cls(field_params, self.grid_xyz)
            self.t0 = ini_field.t0
            a1, a2 = ini_field.get_a12(ini_field.t0)
        else:
            raise NotImplementedError(
                f"`{field_type}` is not implemented, use"
                f"one of {list(SPATIAL_MODEL_FIELDS.keys())}"
            )
        return a1, a2

    def calculate_field(self, t, E_out=None, B_out=None):
        """
        Calculates the electric and magnetic fields at a given time step.
        """
        return super().calculate_field(t, E_out, B_out)



"""
Collection of tests for numerical grid creation.
"""
from copy import deepcopy

import numpy as np
import pytest

from quvac.grid import (
    GridXYZ,
    create_dynamic_grid,
    get_box_size,
    get_bw,
    get_ek,
    get_kmax_grid,
    get_pol_basis,
    get_t_size,
    get_xyz_size,
    setup_grids,
)

TESTED_FIELD_TYPES = ["gaussian", "dipole"]


@pytest.mark.parametrize(
    ("theta", "phi", "expected_ek", "expected_e1", "expected_e2"),
    [
        ( 0, 0,np.array([0,0,1]),np.array([1,0,0]),np.array([0,1,0])),
        (90, 0,np.array([1,0,0]),np.array([0,0,-1]),np.array([0,1,0])),
        (90,90,np.array([0,1,0]),np.array([0,0,-1]),np.array([-1,0,0]))
    ],
)
def test_ek_and_pol_basis(theta, phi, expected_ek, expected_e1, expected_e2):
    theta, phi = np.radians(theta), np.radians(phi)
    ek = get_ek(theta, phi)
    e1, e2 = get_pol_basis(theta, phi)
    assert np.allclose(ek, expected_ek)
    assert np.allclose(e1, expected_e1)
    assert np.allclose(e2, expected_e2)


def get_default_field_params(field_type="gaussian"):
    field_params = {
        "field_type": f"{field_type}_maxwell",
        "lam": 800e-9,
        "tau": 25e-15,
        "theta": 90,
        "phi": 0,
    }
    if field_type == "gaussian":
        field_params["w0"] = 4*800e-9
    return field_params


@pytest.mark.parametrize("field_type", TESTED_FIELD_TYPES)
def test_bandwidth_ok_field_params(field_type):
    field_params = get_default_field_params(field_type)
    bw_perp, bw_long = get_bw(field_params)
    assert bw_perp > 0, "Perpendicular bandwidth should be positive."
    assert bw_long > 0, "Longitudinal bandwidth should be positive."
    # test gaussian with elliptic waist
    if field_type == "gaussian":
        field_params.pop("w0")
        field_params["w0x"] = 4*800e-4
        field_params["w0y"] = 2*800e-4
        bw_perp, bw_long = get_bw(field_params)
        assert bw_perp > 0, "Perpendicular bandwidth should be positive."
        assert bw_long > 0, "Longitudinal bandwidth should be positive."


def test_bandwidth_missing_gauss_params():
    field_params = get_default_field_params("gaussian")
    field_params.pop("w0")
    with pytest.raises(AssertionError):
        get_bw(field_params)


@pytest.mark.parametrize("field_type", TESTED_FIELD_TYPES)
def test_get_kmax_grid(field_type):
    field_params = get_default_field_params(field_type)
    kmax = get_kmax_grid(field_params)
    assert np.all(kmax > 0), "kmax should be positive."
    assert kmax.shape == (3,), "kmax should be a 3-element array."


def test_get_kmax_grid_invalid_field_type():
    field_params = get_default_field_params("invalid_field")
    with pytest.raises(NotImplementedError):
        get_kmax_grid(field_params)


@pytest.mark.parametrize("missing_param", ["lam", "tau", "theta", "phi"])
def test_get_kmax_grid_missing_field_params(missing_param):
    field_params = get_default_field_params("gaussian")
    field_params.pop(missing_param)
    with pytest.raises(AssertionError):
        get_kmax_grid(field_params)


def test_get_t_size():
    Nt = get_t_size(0, 25e-15, lam=800e-9)
    assert Nt > 0, "Nt should be positive."
    assert isinstance(Nt, int), "Nt should be an integer."
    Nt_2 = get_t_size(0, 25e-15, lam=800e-9, grid_res=2)
    assert Nt_2 == 2*Nt-1, "Nt should be double when grid_res is 2."


@pytest.mark.parametrize("field_type", TESTED_FIELD_TYPES)
def test_get_xyz_size(field_type):
    field_params = get_default_field_params(field_type)
    box_size = [10e-6, 10e-6, 10e-6]
    Nxyz = get_xyz_size([field_params], box_size)
    assert isinstance(Nxyz, list), "Nxyz should be a list."
    assert len(Nxyz) == 3, "Nxyz should have 3 elements."

    # test with double grid resolution
    Nxyz_2 = get_xyz_size([field_params], box_size, grid_res=2)
    Nxyz_2_expected_1 = [2*n for n in Nxyz]
    Nxyz_2_expected_2 = deepcopy(Nxyz_2_expected_1)
    Nxyz_2_expected_2[0] -= 1
    condition = (Nxyz_2 == Nxyz_2_expected_1) or (Nxyz_2 == Nxyz_2_expected_2)
    assert condition, "Nxyz should be double when grid_res is 2."

    # test with equal resolution
    Nxyz = get_xyz_size([field_params], box_size, equal_resolution=True)
    assert (Nxyz[0] == Nxyz[1]) and (Nxyz[1] == Nxyz[2]), "grid should be equal in all "
    "dimensions."

    # test with different input format (dict instead of list)
    field_dict = {"field_1": field_params}
    Nxyz = get_xyz_size(field_dict, box_size)
    assert isinstance(Nxyz, list), "Nxyz should be a list."
    assert len(Nxyz) == 3, "Nxyz should have 3 elements."


def get_default_grid_params():
    return {
        "mode": "dynamic",
        "geometry": "x",
        "transverse_factor": 20,
        "longitudinal_factor": 8,
        "time_factor": 2,
    }


@pytest.mark.parametrize("field_type", TESTED_FIELD_TYPES + ["invalid_field"])
def test_get_box_size(field_type):
    field_params = get_default_field_params(field_type)
    grid_params = get_default_grid_params()
    trans, long = get_box_size([field_params], grid_params)
    assert isinstance(trans, float), "Transverse size should be a float."
    assert isinstance(long, float), "Longitudinal size should be a float."


@pytest.mark.parametrize("field_type", TESTED_FIELD_TYPES)
def test_create_dynamic_grid(field_type):
    field_params = get_default_field_params(field_type)
    grid_params = get_default_grid_params()
    grid_params_upd = create_dynamic_grid([field_params], grid_params)
    new_grid_keys = "box_xyz Nxyz box_t Nt".split()
    assert isinstance(grid_params_upd, dict), "grid_params should be a dictionary."
    assert all(key in grid_params_upd for key in new_grid_keys), "grid_params should "
    "contain box_xyz, Nxyz, box_t and Nt keys."

    # test with iterable  grid resolution
    grid_params["spatial_resolution"] = [1, 2, 3]
    grid_params_upd = create_dynamic_grid([field_params], grid_params)
    assert all(key in grid_params_upd for key in new_grid_keys), "grid_params should "
    "contain box_xyz, Nxyz, box_t and Nt keys."


@pytest.fixture
def sample_grid():
    return [
        np.linspace(-5e-6, 5e-6, 200),
        np.linspace(-5e-6, 5e-6, 100),
        np.linspace(-5e-6, 5e-6, 100),
    ]


def test_GridXYZ(sample_grid):
    grid = GridXYZ(sample_grid)
    assert grid.grid_shape == (200, 100, 100), "Grid shape should match input."
    assert grid.grid == sample_grid, "Grid should match input."
    assert grid.kgrid is None, "kgrid should be None by default."

    grid.get_k_grid()
    kgrid_shape = tuple(len(k) for k in grid.kgrid)
    k_keys = "kx ky kz kgrid kgrid_shifted e1x e1y e1z e2x e2y e2z".split()
    assert kgrid_shape == (200, 100, 100), "kgrid shape should match input grid shape."
    assert all(hasattr(grid, key) for key in k_keys), "Grid should contain kgrid keys."


def get_two_default_gaussians():
    field_1 = {
        "field_type": "paraxial_gaussian_maxwell",
        "lam": 800e-9,
        "tau": 25e-15,
        "w0": 4*800e-9,
        "theta": 90,
        "phi": 0,
    }
    field_2 = deepcopy(field_1)
    field_2["phi"] = 180
    return [field_1, field_2]


def compare_grids(fields_1, grid_params_1, fields_2, grid_params_2):
    grid_1, grid_t_1 = setup_grids(fields_1, grid_params_1)
    grid_2, grid_t_2 = setup_grids(fields_2, grid_params_2)

    assert np.allclose(grid_t_1, grid_t_2), "Time grid should match."
    condition = all(np.allclose(g1, g2) for g1, g2 in zip(grid_1.grid, grid_2.grid))  # noqa: B905
    assert condition, "Spatial grids should match."


def test_setup_grids():
    fields = get_two_default_gaussians()
    fields_dict = {
        "field_1": fields[0],
        "field_2": fields[1],
    }
    grid_params = get_default_grid_params()

    compare_grids(fields, grid_params, fields_dict, grid_params)

    grid_params_dynamic = create_dynamic_grid(fields, grid_params)
    box_t = grid_params_dynamic["time_factor"] * fields[0]["tau"]
    grid_params_dynamic["mode"] = "direct"
    grid_params_dynamic["box_t"] = [-box_t, box_t]

    compare_grids(fields, grid_params, fields, grid_params_dynamic)


def test_ignore_idx():
    fields = get_two_default_gaussians()
    # assign very long duration -> unreasonably large time grid if this field is taken
    # taken into account
    fields[1]["tau"] = 100
    grid_params = get_default_grid_params()
    grid_params["ignore_idx"] = [1]
    grid, grid_t = setup_grids(fields, grid_params)





"""
Analytic expression for paraxial gaussian (0-order and higher-orders).

----

.. [1] Y. I. Salamin. "Fields of a Gaussian beam beyond the paraxial 
    approximation." Applied Physics B 86 (2007): 319-326.
.. [2] A. Blinne, et al. "All-optical signatures of quantum vacuum nonlinearities in 
    generic laser fields." PRD 99.1 (2019): 016006 `(article) <https://arxiv.org/abs/1811.08895>`_.
"""

import numexpr as ne
import numpy as np
from scipy.constants import c, pi

from quvac.field.abc import ExplicitField, SpectralField


class GaussianAnalytic(ExplicitField):
    """
    Analytic expression for paraxial Gaussian beam.

    Parameters
    ----------
    field_params : dict
        Dictionary containing the field parameters. Required keys are:
            - 'focus_x' : tuple of float
                Location of spatial focus (x, y, z).
            - 'focus_t' : float
                Location of temporal focus.
            - 'theta' : float
                Polar angle of k-vector (in degrees).
            - 'phi' : float
                Azimuthal angle of k-vector (in degrees).
            - 'beta' : float
                Polarization angle (in degrees).
            - 'lam' : float
                Wavelength of the pulse.
            - 'w0' : float
                Waist size.
            - 'tau' : float
                Duration.
            - 'phase0' : float
                Phase delay at focus.
            - 'E0' : float, optional
                Amplitude (either E0 or W is required).
            - 'W' : float, optional
                Energy (either E0 or W is required).
            - 'alpha_chirp' : float, optional
                Linear chirp in time domain.

    grid : quvac.grid.GridXYZ
        Spatial and grid.

    Notes
    -----
    All field parameters are in SI units.

    Higher-order paraxial Gaussian orders are from [1]_.

    For circular polarization we abide by the following convention:
        - `right-circular` corresponds to (ex + i*ey)
        - `left-circular` to (ex - i*ey)
        
    Difference in field amplitudes for linear and circular polarization should 
    be automatically taken care of by energy correction.
    """

    def __init__(self, field_params, grid):
        super().__init__(field_params, grid)

        self.order = getattr(self, "order", 0)
        self.polarization = getattr(self, "polarization", "linear")

        if "E0" not in field_params:
            err_msg = ("Field params need to have either W (energy) or"
                       "E0 (amplitude) as key")
            assert "W" in field_params, err_msg
            self.E0 = 1.0e10

        # Define additional field variables
        self.x0, self.y0, self.z0 = self.focus_x
        self.t0 = self.focus_t
        self.B0 = self.E0 / c
        self.k = 2.0 * pi / self.lam
        self.omega = c * self.k
        self.zR = pi * self.w0**2 / self.lam

        # Rotate coordinate grid
        self.rotate_coordinates()

        # Define variables not depending on time step
        self.w = "(w0 * sqrt(1. + (z/zR)**2))"
        self.r2 = "(x**2 + y**2)"
        self.R_inv = "z/(z**2 + zR**2)"
        self.E_expr = f"B0 * w0/{self.w} * exp(-{self.r2}/{self.w}**2)"
        self.phase_no_t = ne.evaluate(
            f"phase0 - k*{self.r2}*{self.R_inv}/2. + arctan(z/zR)",
            global_dict=self.__dict__,
        )

        self.E = ne.evaluate(self.E_expr, global_dict=self.__dict__)

        if self.order > 0:
            self.calculate_ho_orders()

        # Set up correct field amplitude
        if "W" in field_params:
            self.check_energy()

    def define_ho_variables(self):
        """
        Define higher order variables.
        """
        self.eps = self.w0 / self.zR
        self.xi = ne.evaluate("x/w0", global_dict=self.__dict__)
        self.nu = ne.evaluate("y/w0", global_dict=self.__dict__)
        self.zeta = ne.evaluate("z/zR", global_dict=self.__dict__)
        self.rho = ne.evaluate("sqrt(xi**2 + nu**2)", global_dict=self.__dict__)
        self.f = ne.evaluate(
            "exp(-1j*arctan(zeta))/sqrt(1 + zeta**2)", global_dict=self.__dict__
        )
        for n in range(1, 5):
            self.__dict__[f"f{n}"] = ne.evaluate(
                "1/(1+zeta**2)**(n/2) * exp(-1j*n*arctan(zeta))",
                global_dict=self.__dict__,
            )

        self.Ex_terms = {
            0: "1",
            2: "eps**2 * f2 * (xi**2 - f1*rho**4/4)",
            4: (
                "eps**4 * f2 * (1/8 - f1*rho**2/4 + f2*rho**2*(xi**2 - rho**2/16) - "
                "f3*rho**4*(xi**2/4 + rho**2/8) + f4*rho**8/32)"
            ),
        }
        self.Ey_terms = {
            0: "0",
            2: "eps**2 * f2",
            4: "eps**4*f4*rho**2 * (1 - f1*rho**2/4)",
        }
        self.Ez_terms = {
            0: "0",
            1: "eps * f1",
            3: "eps**3 * f2 * (-0.5 + f1*rho**2 - f2*rho**4/4)",
            5: (
                "eps**5 * f3 * (-3/8 - 3*f1*rho**2/8 + 17*f2*rho**4/16 - "
                "3*f3*rho**6/8 + f4*rho**8/32)"
            ),
        }
        self.By_terms = {
            0: "1",
            2: "eps**2 * f2*rho**2 * (1/2 - f1*rho**2/4)",
            4: ("eps**4 * f2 * (-1/8 + f1*rho**2/4 + 5*f2*rho**4/16 - f3*rho**6/4 + "
                "f4*rho**8/32)"),
        }
        self.Bz_terms = {
            0: "0",
            1: "eps * f1",
            3: "eps**3*f2 * (1/2 + f1*rho**2/2 - f2*rho**4/4)",
            5: ("eps**5*f3 * (3/8 + 3*f1*rho**2/8 + 3*f2*rho**4/16 - f3*rho**6/4 + "
                "f4*rho**8/32)"),
        }

    def calculate_ho_orders(self):
        """
        Calculate higher order terms for the fields.
        """
        self.define_ho_variables()
        # For a given order, combine final expression to calculate
        names = "Ex Ey Ez By Bz".split()
        for name in names:
            terms = self.__dict__[f"{name}_terms"]
            self.__dict__[f"{name}_expr"] = " + ".join(
                [term for order, term in terms.items() if order <= self.order]
            )
            self.__dict__[f"{name}_ho"] = ne.evaluate(
                self.__dict__[f"{name}_expr"], global_dict=self.__dict__
            )

    def check_energy(self):
        """
        Check and adjust the field energy.
        """
        self._check_energy()

        if self.modify_energy:
            self.E0 *= self.W_correction
            self.B0 = self.E0 / c
            self.E = ne.evaluate(self.E_expr, global_dict=self.__dict__)

    def figure_out_field_components(self, Et):
        if self.polarization == "linear":
            if self.order > 0:
                self.Ex = ne.evaluate("Et * Ex_ho", global_dict=self.__dict__)
                self.Ey = ne.evaluate("Et * Ey_ho * xi * nu",
                                      global_dict=self.__dict__)
                self.Ez = ne.evaluate("-1j*Et * Ez_ho * xi", global_dict=self.__dict__)
                self.Bx = 0.0
                self.By = ne.evaluate("Et * By_ho", global_dict=self.__dict__)
                self.Bz = ne.evaluate("-1j*Et * Bz_ho * nu", global_dict=self.__dict__)
            else:
                self.Ex = self.By = Et.copy()
                self.Ey = self.Ez = self.Bx = self.Bz = 0.0
        elif self.polarization in ["left-circular", "right-circular"]:
            if self.order > 0:
                raise NotImplementedError("Higher paraxial orders for circular "
                                          "polarization are not supported")
            else:
                self.Ex = self.By = Et.copy()
                if self.polarization == "right-circular":
                    self.Ey = 1j * Et.copy()
                else:
                    self.Ey = -1j * Et.copy()
                self.Bx = -self.Ey
                self.Ez = self.Bz = 0.0
        else:
            raise NotImplementedError(f"`{self.polarization}` polarization is not "
                                      "supported")
        
    def get_plane_wave_phase(self, t):
        alpha_chirp = getattr(self, "alpha_chirp", 0)
        plane_phase_expr = "(t-t0-z/c)"
        phase_dict = {
            'c': c,
            't': t, 
            'omega': self.omega,
            'alpha_chirp': alpha_chirp,
        }
        temporal_phase = ne.evaluate(f"(omega*{plane_phase_expr})",
                                     global_dict=self.__dict__, local_dict=phase_dict)
        if alpha_chirp != 0:
            psi_plane_expr = (f"({plane_phase_expr}*omega*"
                              f"(1 + alpha_chirp*{plane_phase_expr}/(tau/2)))")
            psi_plane = ne.evaluate(psi_plane_expr, global_dict=self.__dict__,
                                    local_dict=phase_dict)
        else:
            psi_plane = temporal_phase
        return temporal_phase, psi_plane

    def calculate_field(self, t, E_out=None, B_out=None, mode="real"):
        """
        Calculates the electric and magnetic fields at a given time step.
        """
        k = 2.0 * pi / self.lam # noqa: F841
        self.temporal_phase, self.psi_plane = self.get_plane_wave_phase(t)
        self.phase = "(phase_no_t + psi_plane)"

        Et = ne.evaluate(
            f"E * exp(-(2.*temporal_phase/(omega*tau))**2) * exp(-1.j*{self.phase})",
            global_dict=self.__dict__,
        )

        self.figure_out_field_components(Et)

        if mode == "real":
            self.convert_fields_to_real()

        E_out, B_out = self.rotate_fields_back(E_out, B_out, mode)
        return E_out, B_out


class GaussianEllipticAnalytic(GaussianAnalytic):
    """
    Analytic expression for paraxial elliptical Gaussian beam.

    Parameters
    ----------
    field_params : dict
        Dictionary containing the field parameters. Required keys are:
            - 'focus_x' : tuple of float
                Location of spatial focus (x, y, z).
            - 'focus_t' : float
                Location of temporal focus.
            - 'theta' : float
                Polar angle of k-vector (in degrees).
            - 'phi' : float
                Azimuthal angle of k-vector (in degrees).
            - 'beta' : float
                Polarization angle (in degrees).
            - 'lam' : float
                Wavelength of the pulse.
            - 'w0x', 'w0y' : float
                Waist sizes in x and y directions.
            - 'ellipse_rotation_angle' : float, optional
                Ellipse rotation angle (in degrees).
            - 'tau' : float
                Duration.
            - 'phase0' : float
                Phase delay at focus.
            - 'E0' : float, optional
                Amplitude (either E0 or W is required).
            - 'W' : float, optional
                Energy (either E0 or W is required).
            - 'alpha_chirp' : float, optional
                Linear chirp in time domain.

    grid : quvac.grid.GridXYZ
        Spatial and grid.
    """

    def __init__(self, field_params, grid):
        ExplicitField.__init__(self, field_params, grid)

        self.order = getattr(self, "order", 0)
        self.polarization = getattr(self, "polarization", "linear")

        if "E0" not in field_params:
            err_msg = ("Field params need to have either W (energy) or"
                       "E0 (amplitude) as key")
            assert "W" in field_params, err_msg
            self.E0 = 1.0e10

        # Define additional field variables
        self.x0, self.y0, self.z0 = self.focus_x
        self.t0 = self.focus_t
        self.B0 = self.E0 / c
        self.k = 2.0 * pi / self.lam
        self.omega = c * self.k
        self.zRx = pi * self.w0x**2 / self.lam
        self.zRy = pi * self.w0y**2 / self.lam

        # Rotate coordinate grid
        self.rotate_coordinates()

        # Account for possible ellipse rotation
        self.account_for_ellipse_rotation()

        # Define variables not depending on time step
        self.wx = "(w0x * sqrt(1. + (z/zRx)**2))"
        self.wy = "(w0y * sqrt(1. + (z/zRy)**2))"
        self.Rx_inv = "z/(z**2 + zRx**2)"
        self.Ry_inv = "z/(z**2 + zRy**2)"
        self.w_prefactor = "sqrt(zRx*zRy / (sqrt(z**2 + zRx**2) * sqrt(z**2 + zRy**2)))"
        self.focusing = f"exp(-(xrot/{self.wx})**2 - (yrot/{self.wy})**2)"
        self.E_expr = f"B0 * {self.w_prefactor} * {self.focusing}"
        self.phaseg = "(arctan(z/zRx) + arctan(z/zRy))/2"
        self.phase_curved = f"k/2.*(xrot**2*{self.Rx_inv} + yrot**2*{self.Ry_inv})"
        self.phase_no_t = ne.evaluate(
            f"phase0 - {self.phase_curved} + {self.phaseg}",
            global_dict=self.__dict__,
        )

        self.E = ne.evaluate(self.E_expr, global_dict=self.__dict__)

        # Set up correct field amplitude
        if "W" in field_params:
            self.check_energy()

    def account_for_ellipse_rotation(self):
        self.ellipse_rotation_angle = getattr(self, "ellipse_rotation_angle", 0)
        if self.ellipse_rotation_angle != 0:
            ell0 = self.ellipse_rotation_angle
            self.xrot = self.x*np.cos(ell0) + self.y*np.sin(ell0)
            self.yrot = -self.x*np.sin(ell0) + self.y*np.cos(ell0)
        else:
            self.xrot, self.yrot = self.x, self.y


class GaussianSpectral(SpectralField):
    """
    Analytic expression for paraxial Gaussian beam.

    Parameters
    ----------
    field_params : dict
        Dictionary containing the field parameters. Required keys are:
            - 'focus_x' : tuple of float
                Location of spatial focus (x, y, z).
            - 'focus_t' : float
                Location of temporal focus.
            - 'theta' : float
                Polar angle of k-vector (in degrees).
            - 'phi' : float
                Azimuthal angle of k-vector (in degrees).
            - 'beta' : float
                Polarization angle (in degrees).
            - 'lam' : float
                Wavelength of the pulse.
            - 'w0' : float
                Waist size.
            - 'tau' : float
                Duration.
            - 'phase0' : float
                Phase delay at focus.
            - 'E0' : float, optional
                Amplitude (either E0 or W is required).
            - 'W' : float, optional
                Energy (either E0 or W is required).
            - 'alpha_chirp' : float, optional
                Linear chirp in frequency domain.

    grid : quvac.grid.GridXYZ
        Spatial and grid.

    Notes
    -----
    All field parameters are in SI units.

    The expression for the spectrum is taken from [2]_ (Eq. 24). We write it here
    assuming the central k-vector is along kz. 
    """

    def __init__(self, field_params, grid):
        super().__init__(field_params, grid)

        if "E0" not in field_params:
            err_msg = ("Field params need to have either W (energy) or"
                       "E0 (amplitude) as key")
            assert "W" in field_params, err_msg
            self.E0 = 1.0e10

        # Define additional field variables
        self.x0, self.y0, self.z0 = self.focus_x
        self.t0 = self.focus_t
        self.B0 = self.E0 / c
        self.k = 2.0 * pi / self.lam
        self.omega = c * self.k
        self.alpha_chirp = getattr(self, "alpha_chirp", 0)

        # Rotate k-space grid
        self.rotate_kgrid()

        self.kperp2 = "(kx**2 + ky**2)"
        self.define_vector_potential_expression()
        self.vector_potential = ne.evaluate(self.vector_potential_expr,
                                            local_dict=self.vector_potential_dict)
        self.vector_potential = np.fft.ifftshift(self.vector_potential)
        # Since the spectrum expression is given for continuous Fourier transform
        # and in the code I use DFT, there is a phase factor related to the start
        # of the spatial grid that relates two transforms.
        self.vector_potential *= self.dft_factor
        
        self.Ax, self.Ay, self.Az = self.rotate_vector_potential_back()

    def define_vector_potential_expression(self):
        self.vector_potential_expr = (
            "where(kz > 0, pi**1.5/2 * 1j/(1j*kabs) * kz/kabs * E0*tau*w0**2 * "
            f"exp(-(w0/2)**2*{self.kperp2}) * "
            "exp(-(tau/4)**2*(c*kabs-omega)**2*(1-1j*alpha)) * "
            "exp(-1j*(x0*kx + y0*ky + z0*kz - phase0 - c*kabs*t0)), 0)"
        )
        self.vector_potential_dict = {
            "pi": pi,
            "c": c,
            "kx": self.kx_rotated,
            "ky": self.ky_rotated,
            "kz": self.kz_rotated,
            "kabs": self.kabs_rotated,
            "E0": self.E0,
            "tau": self.tau,
            "w0": self.w0,
            "omega": self.omega,
            "alpha": self.alpha_chirp,
            "x0": self.x0,
            "y0": self.y0,
            "z0": self.z0,
            "phase0": self.phase0,
            "t0": self.t0,
        }
        x0, y0, z0 = [ax[0] for ax in self.grid]
        kx, ky, kz = self.kmeshgrid
        self.dft_factor = np.exp(1j*(kx*x0 + ky*y0 + kz*z0))
    
    def calculate_field(self, t, E_out=None, B_out=None, mode="real"):
        """
        Calculates the electric and magnetic fields at a given time step.
        """
        raise NotImplementedError("GaussianSpectral works only as a model field" \
        "for MaxwellField")
    


"""
Abstract interfaces for different integrators.

.. warning::
    Currently not used...
"""

from abc import ABC, abstractmethod


class Integrator(ABC):
    """
    Abstract integrator class.
    """
    @abstractmethod
    def calculate_amplitudes(self, t_grid):
        """ """
        ...
